/** 
 * @Version $Rev: 200 $
 * @Author M.Sekido
 * @Date  $Date: 2016-08-28 22:40:41 +0900 (Sun, 28 Aug 2016) $
 * $Id: InVSI_mdl.cpp 200 2016-08-28 13:40:41Z vlbi $ 
 */

#ifndef LINUX
#define LINUX
#endif
#ifndef _LARGEFILE_SOURCE 
#define _LARGEFILE_SOURCE
#endif 
#define _LARGEFILE64_SOURCE

/** Input Device (K5_VSI) Model Class Library 
 * @Version $Rev: 200 $
 * @Author M.Sekido
 * @Date  $Date: 2016-08-28 22:40:41 +0900 (Sun, 28 Aug 2016) $
 * $Id: InVSI_mdl.cpp 200 2016-08-28 13:40:41Z vlbi $ .
 */

#include "InVSI_mdl.hpp"
int K5_VSI::device = 0;
int K5_VSI::mem = 0;
int K5_VSI::num_opened = 0;
Mutex K5_VSI::k5vsi_mutex;
int K5_VSI::frame_epoch_res_msec = FRAME_EPOCH_RES_MSEC; // msec
string K5_VSI::vsi_devname("/dev/k5vsi");
string K5_VSI::mem_devname("/dev/mem");
int K5_VSI::npage =0;
Tm_mem  K5_VSI::device_epoch; ///< Epoch obtained from K5-VSI
uint32_t K5_VSI::page_length[MAX_K5VSI_PAGES]; /// pagesize in word(4byte)
uint32_t K5_VSI::page_offset[MAX_K5VSI_PAGES]; // 
bool K5_VSI::page_offset_setflag[MAX_K5VSI_PAGES]; // 

/** Constructor */
K5_VSI::K5_VSI():
  //  device(0), mem(0), 
  //  device_epoch(0),
  //  vsi_devname("/dev/k5vsi"), mem_devname("/dev/mem"),
  // npage(0), 
  page_ptr(0),
  rpage(0), map_enable(false), dev_opened_flag(false)
{
  //  make_tab(enc_table); 
  
  SetID("K5_VSI");
  vlbi_frame = new K5VSI_Frame;
  bzero((void*)page_length, MAX_K5VSI_PAGES*sizeof(uint32_t));
  for(int i=0; i< MAX_K5VSI_PAGES; i++){
    page_offset[i] =0;
    page_offset_setflag[i] = false;
  }
}

/** Destructor */
K5_VSI::~K5_VSI(){
  ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF); // add 2014/11/17 miya
  if(device!=0) CloseDev();
  //  if( vlbi_frame != 0){ delete vlbi_frame; vlbi_frame=0;}
}

const int K5_VSI::CloseDev(){

  num_opened --;
  dev_opened_flag=false;

  Logsystem::info_logger()->Tstamp()
      << "I K5_VSI::CloseDev() " << num_opened << " thds still opened\n"; 

  if(num_opened >0) return 0; // Still some others are working

  if(device != 0) close(device);
  if(mem != 0) close(mem);
  device = 0;
  mem=0;
  Logsystem::info_logger()->Tstamp() << "D K5_VSI: CloseDev().."; 
//   k5vsi_mutex.lock();
//   ioctl(device,SET_FRAME_FREEZE,0x0);
//   k5vsi_mutex.unlock();
  *Logsystem::info_logger() << "--CloseDev()..OK\n"; 
  Logsystem::info_logger()->Flush();
  return 0;
}

const int K5_VSI::TestDev() {

  try{
    char buf[2048];
    k5vsi_mutex.lock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::TestDev(): mutex locked.\n";
    if((device=open(vsi_devname.c_str(), O_RDWR))==-1) {
      sprintf(buf,"E K5_VSI::TestDev(): device file [%s] open error\n",
	      vsi_devname.c_str());
      throw buf;
      //return -1;
    }
    close(device);
    if(( mem=open(mem_devname.c_str() ,O_RDONLY))==-1) {
      sprintf(buf,"device file [%s] open error, %s\n",
	      mem_devname.c_str(),strerror(errno));
      throw buf;
      //    return -1;
    }
    close(mem);
    k5vsi_mutex.unlock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::TestDev(): mutex UNlocked.\n";
    return 0;
  }
  
  catch (char* str){
    fprintf(stderr,"%s \n", str);
    Logsystem::error_logger()->Tstamp()  << str << "\n";   
    k5vsi_mutex.unlock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::TestDev(): mutex UNlocked.\n";
    return -1;
  }

}


const int K5_VSI::OpenDev(const InDev_Model::DevCheckLevel_t dclevel){

  try{
    char buf[1024];

    k5vsi_mutex.lock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex locked.\n";
    if(dev_opened_flag){
      cerr << "I K5_VSI::OpenDev(): is already Opended" << endl;
      //     k5vsi_mutex.lock();
      //     if( ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF) !=0) {
      //       Logsystem::info_logger()->Tstamp() 
      // 	<< "W K5_VSI::DevOpen(): non-stopping DMA ioctl(SET_FRAME_FREEZE) failed.\n";
      //     }
      if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
	Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
    k5vsi_mutex.unlock();
      return 0;   // Dev is already opened nothing to do.
    }
    
    if(num_opened == 0){ // Firstly opened 
      cerr << "I K5_VSI::OpenDev(): The First Opening K5_VSI[" << vsi_devname << "]" << endl;
      if((device=open(vsi_devname.c_str(), O_RDWR))==-1) {
	sprintf(buf,"E K5_VSI::OpenDev(): VSI-device file [%s] open error\n",
		vsi_devname.c_str());
	if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
	  Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
	k5vsi_mutex.unlock();
	throw buf;
      }

      cerr << "I K5_VSI::OpenDev(): The First Opening Mem[" << mem_devname << "]" << endl;
      if(( mem=open(mem_devname.c_str() ,O_RDONLY))==-1) {
	sprintf(buf,"E K5_VSI::OpenDev():MeM-device file [%s] open error, %s\n",
		mem_devname.c_str(), strerror(errno));
	if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
	  Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
	k5vsi_mutex.unlock();
	throw buf;
      }

      Logsystem::info_logger()->Tstamp() << "K5_VSI Opened \n";
      Logsystem::info_logger()->Flush();

      Logsystem::debug_logger()->Tstamp() << "D K5_VSI: Set_Frame_Freeze.."; 
      int ioret = ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF);
      usleep(INTVL_IOCTL_USEC);
      *Logsystem::debug_logger() << "--Set_Frame_Freeze..end\n"; 
      Logsystem::debug_logger()->Flush();
      if(ioret != 0){
	Logsystem::warning_logger()->Tstamp() 
	  << "Warning in DevOpen(): non-stopping DMA ioctl(SET_FRAME_FREEZE) failed.\n";
      }
      
      Logsystem::debug_logger()->Tstamp() << "D K5_VSI: GetFrameNumber.."; 
      npage = ioctl(device,GET_FRAME_NUMBER,0x00); //miya null->0x00
      usleep(INTVL_IOCTL_USEC);
      *Logsystem::debug_logger() << "--GetFrameNumber..OK\n"; 
      Logsystem::debug_logger()->Flush();
      if(npage > MAX_K5VSI_PAGES ){
	sprintf(buf,"E K5_VSI:GET_FRAME_NUMBER failed");
	if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
	  Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
	k5vsi_mutex.unlock();
	throw buf;
      }
      ulong Clock = (uint32_t)ioctl(device,GET_FRAME_CLOCK ,NULL);//miya 0->NULL
      usleep(INTVL_IOCTL_USEC);
      ulong exp_clock = vlbi_frame->DAS_Inf().Bytes_sec()/4;
      if( exp_clock != Clock){
	sprintf(buf,"E K5_VSI::OpenDev():Expected Clock(%lu) != Board Clock(%lu)",
		exp_clock, Clock);
	Logsystem::error_logger()->Tstamp() << buf << "\n";
	if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
	  Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
	k5vsi_mutex.unlock();
	if(dclevel == InDev_Model::STRICT){ // In case of Strict-Check, break, else OK
	  throw buf;
	}
      }
      
    } // end if the firs time 

    Logsystem::info_logger()->Tstamp()
      << "I K5_VSI::OpenDev() " << num_opened << " thds opened\n"; 
    sprintf(buf,"K5_VSI-%d thd", num_opened);
    SetID(buf);
    dev_opened_flag = true;
    num_opened++; // Count-up 
    // by M.S.2013.6.16 to avoid double opneing of Devicefile
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::OpenDev(): mutex UNlocked.\n";
    k5vsi_mutex.unlock();

    //    Logsystem::debug_logger()->Tstamp() << "D K5_VSI: GetDeviceEpoch().."; 
    //    time_t dev_epoch_now = ioctl(device,GET_DEVICE_EPOCH,NULL);
    time_t dev_epoch_now = GetDevEpoch();
    //    *Logsystem::debug_logger() << "--GetDevEpoch()..OK\n"; 
    //    Logsystem::debug_logger()->Flush();

    if(dev_epoch_now ==-1){
      strcpy(buf, "E K5_VSI::OpenDev(): GET_DEVICE_EPOCH failed.");
      throw buf;
    }
    device_epoch.Set_time(dev_epoch_now);
    
    char tbuf[1024];
    strftime(tbuf,80,"%Y/%j-%H:%M:%S", gmtime(&dev_epoch_now));

    struct timeval st;
    gettimeofday(&st,NULL);
    int tdif = abs(st.tv_sec - dev_epoch_now);
    if( tdif > 10){
      sprintf(buf, "E K5_VSI::OpenDev(): device time(%s) = %lu differs from PC time(%lu) by %d sec.",
	      tbuf, (unsigned long)dev_epoch_now, (unsigned long)st.tv_sec, tdif);
      throw buf;
//       cerr <<  "E K5_VSI::OpenDev(): device time(" << buf
// 	   << " =" << device_epoch << ") differs from PC time("
// 	   << st.tv_sec << ") by " << tdif << "sec\n";
//       Logsystem::error_logger()->Tstamp() 
// 	<<  "E K5_VSI::OpenDev(): device time(" << buf
// 	<< " =" << device_epoch << ") differs from PC time("
// 	<< st.tv_sec << ") by " << tdif << "sec\n";
//       return -1;
    }
    fprintf(stderr,"\tGet_Device time= %s at Device Openning.\n", tbuf);
    Logsystem::info_logger()->Tstamp() 
      << "\tGet_Device time= " << tbuf << " at Device Openning.\n";
    
    map_enable = true;
    return 0;
  }

  catch (char* str){
    fprintf(stderr,"%s \n", str);
    switch(str[0]){
    case 'D':
      Logsystem::debug_logger()->Tstamp()  << str << "\n";  break; 
    case 'E':
      cerr << str << endl;
      Logsystem::error_logger()->Tstamp()  << str << "\n";  break; 
    case 'I':
    default:
      Logsystem::info_logger()->Tstamp()  << str << "\n";  break; 
    }
    return -1;  
}

  catch (string str){
    fprintf(stderr,"%s \n", str.c_str());
    //    Logsystem::error_logger()->Tstamp()  << str << "\n";   
    switch(str[0]){
    case 'D':
      Logsystem::debug_logger()->Tstamp()  << str << "\n";  break; 
    case 'E':
      Logsystem::error_logger()->Tstamp()  << str << "\n";  break; 
    case 'I':
    default:
      Logsystem::info_logger()->Tstamp()  << str << "\n";  break; 
    }
    
    return -1;
  }
  

}


const int32_t K5_VSI::Srate()const {
  if( vlbi_frame == 0  ){
    cerr << "E K5_VSI::Srate(): *vlbi_frame is empty." << endl;
    Logsystem::error_logger()->Tstamp() 
      << "E K5_VSI::Srate(): *vlbi_frame is empty.\n";
    return -1;
  }
  return vlbi_frame->DAS_Inf().Srate();
} 


const int K5_VSI::Nch()const{
  if( vlbi_frame == 0 ){
    cerr << "E K5_VSI::Nch(): *vlbi_frame is empty." << endl;
    Logsystem::error_logger()->Tstamp() 
      << "E K5_VSI::Nch(): *vlbi_frame is empty.\n";
    return -1;
  }
  return vlbi_frame->DAS_Inf().Nch();
} 


const int K5_VSI::Nbit()const {
  if( vlbi_frame == 0){
    cerr << "E K5_VSI::Nbit(): *vlbi_frame is empty." << endl;
    Logsystem::error_logger()->Tstamp() 
      << "E K5_VSI::Nbit(): *vlbi_frame is empty.\n";
    return -1;
  }
  return vlbi_frame->DAS_Inf().Qbit();
}

const int K5_VSI::Set_Mode(const char* a){
  if( vlbi_frame == 0 ){
    cerr << "E K5_VSI::Set_Mode(): *vlbi_frame is empty." << endl;
    Logsystem::error_logger()->Tstamp() 
      << "E K5_VSI::Set_Mode(): *vlbi_frame is empty.\n";
    return -1;
  }
  int32_t srate(0);
  int nch(0);
  string tmp =a;
  int okflag=0;
  if(tmp.find("64Msps/2bit/16ch")!= tmp.npos) {srate=64000000;nch=16;okflag=1;}
  else if(tmp.find("32Msps/2bit/16ch")!= tmp.npos) {srate=32000000;nch=16;okflag=1;}
  else if(tmp.find("16Msps/2bit/16ch")!= tmp.npos) {srate=16000000;nch=16;okflag=1;}
  else if(tmp.find("08Msps/2bit/16ch")!= tmp.npos) {srate= 8000000;nch=16;okflag=1;}
  else if(tmp.find("04Msps/2bit/16ch")!= tmp.npos) {srate= 4000000;nch=16;okflag=1;}
  else if(tmp.find("64Msps/2bit/08ch")!= tmp.npos) {srate=64000000;nch=8;okflag=1;}
  else if(tmp.find("32Msps/2bit/08ch")!= tmp.npos) {srate=32000000;nch=8;okflag=1;}
  else if(tmp.find("16Msps/2bit/08ch")!= tmp.npos) {srate=16000000;nch=8;okflag=1;}
  else if(tmp.find("08Msps/2bit/08ch")!= tmp.npos) {srate= 8000000;nch=8;okflag=1;}
  else if(tmp.find("04Msps/2bit/08ch")!= tmp.npos) {srate= 4000000;nch=8;okflag=1;}
  else if(tmp.find("64Msps/2bit/04ch")!= tmp.npos) {srate=64000000;nch=8;okflag=1;}
  else if(tmp.find("32Msps/2bit/04ch")!= tmp.npos) {srate=32000000;nch=4;okflag=1;}
  else if(tmp.find("16Msps/2bit/04ch")!= tmp.npos) {srate=16000000;nch=4;okflag=1;}
  else if(tmp.find("08Msps/2bit/04ch")!= tmp.npos) {srate= 8000000;nch=4;okflag=1;}
  else if(tmp.find("04Msps/2bit/04ch")!= tmp.npos) {srate= 4000000;nch=4;okflag=1;}
  else if(tmp.find("64Msps/2bit/02ch")!= tmp.npos) {srate= 64000000;nch=2;okflag=1;}
  else if(tmp.find("32Msps/2bit/02ch")!= tmp.npos) {srate= 32000000;nch=2;okflag=1;}
  else if(tmp.find("16Msps/2bit/02ch")!= tmp.npos) {srate= 16000000;nch=2;okflag=1;}
  else if(tmp.find("08Msps/2bit/02ch")!= tmp.npos) {srate= 8000000;nch=2;okflag=1;}
  else if(tmp.find("04Msps/2bit/02ch")!= tmp.npos) {srate= 4000000;nch=2;okflag=1;}
  else{    okflag=0; nch=0; srate=0;  }

//   if(tmp.find("64Msps/2bit/16ch")!= tmp.npos) {speed=64000000;srate=64000000;nch=16;okflag=1;}
//   if(tmp.find("32Msps/2bit/16ch")!= tmp.npos) {speed=32000000;srate=32000000;nch=16;okflag=1;}
//   if(tmp.find("16Msps/2bit/16ch")!= tmp.npos) {speed=16000000;srate=16000000;nch=16;okflag=1;}
//   if(tmp.find("08Msps/2bit/16ch")!= tmp.npos) {speed= 8000000;srate= 8000000;nch=16;okflag=1;}
//   if(tmp.find("04Msps/2bit/16ch")!= tmp.npos) {speed= 4000000;srate= 4000000;nch=16;okflag=1;}
//   if(tmp.find("64Msps/2bit/08ch")!= tmp.npos) {speed=32000000;srate=64000000;nch=8;okflag=1;}
//   if(tmp.find("32Msps/2bit/08ch")!= tmp.npos) {speed=16000000;srate=32000000;nch=8;okflag=1;}
//   if(tmp.find("16Msps/2bit/08ch")!= tmp.npos) {speed= 8000000;srate=16000000;nch=8;okflag=1;}
//   if(tmp.find("08Msps/2bit/08ch")!= tmp.npos) {speed= 4000000;srate= 8000000;nch=8;okflag=1;}
//   if(tmp.find("04Msps/2bit/08ch")!= tmp.npos) {speed= 2000000;srate= 4000000;nch=8;okflag=1;}
//   if(tmp.find("64Msps/2bit/04ch")!= tmp.npos) {speed=16000000;srate=64000000;nch=8;okflag=1;}
//   if(tmp.find("32Msps/2bit/04ch")!= tmp.npos) {speed= 8000000;srate=32000000;nch=4;okflag=1;}
//   if(tmp.find("16Msps/2bit/04ch")!= tmp.npos) {speed= 4000000;srate=16000000;nch=4;okflag=1;}
//   if(tmp.find("08Msps/2bit/04ch")!= tmp.npos) {speed= 2000000;srate= 8000000;nch=4;okflag=1;}
//   if(tmp.find("04Msps/2bit/04ch")!= tmp.npos) {speed= 1000000;srate= 4000000;nch=4;okflag=1;}
//   if(tmp.find("64Msps/2bit/02ch")!= tmp.npos) {speed= 8000000;srate= 64000000;nch=2;okflag=1;}
//   if(tmp.find("32Msps/2bit/02ch")!= tmp.npos) {speed= 4000000;srate= 32000000;nch=2;okflag=1;}
//   if(tmp.find("16Msps/2bit/02ch")!= tmp.npos) {speed= 2000000;srate= 16000000;nch=2;okflag=1;}
//   if(tmp.find("08Msps/2bit/02ch")!= tmp.npos) {speed= 1000000;srate= 8000000;nch=2;okflag=1;}
//   if(tmp.find("04Msps/2bit/02ch")!= tmp.npos) {speed=  500000;srate= 4000000;nch=2;okflag=1;}

  if(okflag){
    vlbi_frame->Set_Sparam(nch,2,srate);
    return 0;
  }else{
    cerr << "W K5_VSI::Set_Mode(): Bad mode option " << tmp << endl;
    Logsystem::info_logger()->Tstamp() 
      << "W K5_VSI::Set_Mode(): Bad mode option " << tmp << "\n";
    return -1;
  }
  
}

string K5_VSI::Mode(){
  if( vlbi_frame == 0 ){
    cerr << "E K5_VSI::Set_Nbit(): *vlbi_frame is empty." << endl;
    Logsystem::info_logger()->Tstamp() 
      << "E K5_VSI::Set_Nbit(): *vlbi_frame is empty.\n";
    return "";
  }
  return vlbi_frame->DAS_Inf().Mode();
}


/// Mapping data array to the memory
const uint8_t* K5_VSI::MappingData(time_t epoch){

  if(device ==0){
    cerr << "Error @K5_VSI::MappingData("<< Thread::StrID() << "): K5VSI Device is not yet opened\n";
    Logsystem::error_logger()->Tstamp()
      << "Error @K5_VSI::MappingData("<< Thread::StrID() << "): K5VSI Device is not yet opened\n";
    return 0;
  }
  if(! map_enable){
    cerr << "Warning @K5_VSI::MappingData("<< Thread::StrID() << "): Map Disabled\n";
    Logsystem::info_logger()->Tstamp()
      << "Warning @K5_VSI::MappingData("<< Thread::StrID() << "): Map Disabled\n";
    return 0;
  }

  time_t t_now(0);
  rpage = epoch % npage;
  bool found = false;
  bool device_ep = false; // add 2014/10/27 miya
  char time_code[3][80]; // miya [2]->[3]
  char buf[1024];
  time_t t_now_last(0);
  const int count_limit=12;
  int waiting_count= count_limit;

  while(found == false && waiting_count> 0){
    //    k5vsi_mutex.lock();
    //    t_now = ioctl(device,GET_FRAME_EPOCH, rpage);
    Logsystem::debug_logger()->Tstamp() << "D K5_VSI::Mapping("<< Thread::StrID() 
					<< "): Calling GetFrameEpoch(" << rpage << ")\n";
  	//ioctl(device,SET_FRAME_FREEZE,epoch); //comment 2014/10/27 miya
  	//while((epoch+2)> ioctl(device,GET_DEVICE_EPOCH,0)) usleep(10000); //comment 2014/10/27 miya
  	// add 2014/10/27 miya start-----------------------------------------
  	time_t epoch_tmp;
	while((epoch+2) > (epoch_tmp =ioctl(device,GET_DEVICE_EPOCH,0))){
		if (! device_ep){
			strftime(time_code[2],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&epoch_tmp));
		Logsystem::debug_logger()->Tstamp()
		<< "D " << " InVSI_mdl::MappingData[" << Thread::StrID()
		<< "] deviceEpoch (" << time_code[2] << "\n"; 
		Logsystem::debug_logger()->Flush();
		//printf("epoch=%lu, deviceEpoch=%lu\n", epoch, epoch_tmp); 
		}
	usleep(10000);
	device_ep=true;
	}// add 2014/10/27 miya end-----------------------------------------
  	//while(1){ k5vsi_mutex.lock(); t_now = GetFrameEpoch(rpage); k5vsi_mutex.unlock();}
  	t_now = GetFrameEpoch(rpage);
	// MS added 2014/9/1  --- from here comment miya 2014/11/17
  	//ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF);
	//usleep(INTVL_IOCTL_USEC);
	// MS added 2014/9/1 -- to here end miya 2014/11/17

    //    k5vsi_mutex.unlock();
    
     //    fprintf(stderr, "Mapping:epoch=%lu t_now=%lu\n", epoch, t_now);
    //     Logsystem::debug_logger()->Tstamp() 
    //       << "Mapping:epoch=" << epoch << " t_now=" << t_now << "\n";
    if(Logsystem::Test_Logsys(Logsystem::DEBUG)){
      strftime(time_code[0],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&epoch));
      Logsystem::debug_logger()->Tstamp()
	<< "D " << " InVSI_mdl::MappingData[" << Thread::StrID()
	<< "] Expecting (" << epoch << "," << time_code[0] << ") to read page("
	<< rpage << ")=" << t_now << "\n"; 
	// << StrPageState ( 1 ) << "\n";
      Logsystem::debug_logger()->Flush();
    }      
    if(t_now < epoch){
      int tdif = epoch - t_now;
      strftime(time_code[0],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&epoch));
      strftime(time_code[1],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&t_now));
      if(tdif > 10){
	sprintf(buf,
		"Warning in K5_VSI::MappingData(): expecting epoch(%s) >tnow(%s) more than 10sec.",
		time_code[0], time_code[1]);
	fprintf(stderr,"%s\n", buf);
	Logsystem::debug_logger()->Tstamp() << buf << "\n";
	Logsystem::debug_logger()->Flush();
      }
      
      sprintf(buf,"Waiting in K5VSI::MappingData() %s --now=%s, Tdif= %d secn", 
	      time_code[0], time_code[1], tdif);
      fprintf(stderr,"%s\n", buf);
      Logsystem::info_logger()->Tstamp() << buf << "\n";
      Logsystem::info_logger()->Flush();      

      if(t_now_last == t_now)      waiting_count --;
      t_now_last = t_now;
      usleep(1000000);
      continue;
    } // end of wait for future.
    
    if(t_now > epoch ){ // Requested Time is already past 
      strftime(time_code[0],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&epoch));
      strftime(time_code[1],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&t_now));
      cerr << "Error in K5_VSI::MappingData(): expecting epoch("
	   << time_code[0] <<  ")  is already past dev-tnow(" << time_code[1] << ")\n";
      Logsystem::error_logger()->Tstamp() 
	<< "Error in K5_VSI::MappingData(): expecting epoch("
	<< time_code[0] <<  ")  is already past dev-tnow(" << time_code[1] << ")\n";
      Logsystem::debug_logger()->Tstamp() 
	<< "Error in K5_VSI::MappingData(): expecting epoch("
	<< time_code[0] <<  ")  is already past dev-tnow(" << time_code[1] << ")\n";
      Logsystem::debug_logger()->Flush();
      //return 0; comment 2014/11/11 miya
    	break; // add 2014/11/11 miya
    } 

    Logsystem::debug_logger()->Tstamp() << "D K5_VSI::Mapping(): Epoch(" << t_now << ") Mached.\n";

    found = true;
    ioctl(device,SET_FRAME_FREEZE,epoch); // add miya 2014/10/27
  } // End of while loop to find the requested time data.

  page_ptr=0; //add 2015/2/13 miya
  if(found){
    //    k5vsi_mutex.lock();
    // page_length is page_size(1sec) in 4-byte unit 
    //    Logsystem::debug_logger()->Tstamp() << "D K5_VSI: GetFrameClock.."; 
    
    //    uint32_t page_length=(unsigned int)ioctl(device,GET_FRAME_CLOCK ,rpage);
    size_t pagesize = sizeof(unsigned int) * PageLength(rpage);
    //    *Logsystem::debug_logger() << "--GetFrameClock..OK\n"; 

    // Get top address of that page
    //    Logsystem::debug_logger()->Tstamp() << "D K5_VSI: GetFrameOffset.."; 
    //    uint32_t page_offset=(unsigned int)ioctl(device,GET_FRAME_OFFSET,rpage);
    uint32_t page_offsetr = GetPageOffset(rpage);
    //    *Logsystem::debug_logger() << "--GetFrameOffset..OK\n"; 
    //    Logsystem::debug_logger()->Flush();
    //    k5vsi_mutex.unlock();  


    // Mapping page to memory
    //  page_ptr =(uint8_t *)mmap(0x0,pagesize, PROT_READ,MAP_SHARED,mem,page_offset);
    page_ptr =(uint8_t *)mmap(0x0,pagesize, PROT_READ,MAP_PRIVATE,mem,page_offsetr);

  	if(Logsystem::Test_Logsys(Logsystem::INFO)){
      char buf[1024];
      int now_epoch = time(0) - epoch;
      strftime(time_code[0],sizeof(time_code),"%Y/%j %H:%M:%S", gmtime(&epoch));
      sprintf(buf,"I K5_VSI::Mapping[Epoch=%lu,%s] page[%d]Offset=%u size=%lu, addr= %0lX, epc-now=%d\n", 
	      (unsigned long)epoch, time_code[0], rpage, page_offsetr, 
	      (unsigned long)pagesize, (unsigned long)page_ptr, now_epoch);
      Logsystem::info_logger()->Tstamp() << buf ;
      Logsystem::info_logger()->Flush();
    }
  }else{ // If not found.
    Logsystem::error_logger()->Tstamp()
      << "E K5_VSI::MappingData(): VSI card Epoch was not updated. Card problem is suggested.\n"; 
    cerr << "******************************************************\n";
    cerr << "E K5_VSI::MappingData(): VSI card Epoch was not updated. Card problem is suggested.\n";
    cerr << "******************************************************" << endl;
  }

  return page_ptr;
  
}


// // No inversion
// int make_tab_ni(unsigned char table[256])
// {

//   int index;
//   memset(table,0,256);
//   for(index=0;index<256;index++){table[index]= index;  }

//   fprintf(stderr,"Bit encoding No change\n");
//   return 0;
// }

// /// Inverting 2 bit
// /// This makes nomal offset-binary codeing, because  
// /// ADS-2000 output msb/lsb is opposite it its 2 bit sampling.
// int make_tab(unsigned char table[256])
// {
//   int index;
//   fprintf(stderr,"Bit encoding OffsetBin\n");
//   memset(table,0,256);
//   for(index=0;index<256;index++){
//     if(index&0x80) table[index]|=0x40;
//     if(index&0x40) table[index]|=0x80;
//     if(index&0x20) table[index]|=0x10;
//     if(index&0x10) table[index]|=0x20;
//     if(index&0x08) table[index]|=0x04;
//     if(index&0x04) table[index]|=0x08;
//     if(index&0x02) table[index]|=0x01;
//     if(index&0x01) table[index]|=0x02;
//   }
//   return 0;
// }

// int make_tab_sgnmag(unsigned char table[256])
// {
//   int index;
//   fprintf(stderr,"Bit encoding SignManitude\n");

//   memset(table,0,256);

//   for(index=0;index<256;index++){
//     if(index&0x80) table[index]|=0x40;
//     if(index&0x40) table[index]|=0x80;
//     if(index&0x20) table[index]|=0x10;
//     if(index&0x10) table[index]|=0x20;
//     if(index&0x08) table[index]|=0x04;
//     if(index&0x04) table[index]|=0x08;
//     if(index&0x02) table[index]|=0x01;
//     if(index&0x01) table[index]|=0x02;
//   }
  
//   for(int i=0;i<256;i++){
//     if( (table[i] & 0x80) == 0) table[i] ^= 0x40; 
//     if( (table[i] & 0x20) == 0) table[i] ^= 0x10; 
//     if( (table[i] & 0x08) == 0) table[i] ^= 0x04; 
//     if( (table[i] & 0x02) == 0) table[i] ^= 0x01; 
//   }
  
//   return 0;
// }

// int32_t K5_VSI::Convert(vector <uint8_t> & arr, int iofst){

//   //  int32_t rq_Byte_sec = srate/8*nch*nbit;
//   int32_t rq_Byte_sec = vlbi_frame->DAS_Inf().Speed()*4;
//   int32_t srate = vlbi_frame->DAS_Inf().Srate();
//   if( arr.size() < (uint32_t)rq_Byte_sec ) arr.resize(rq_Byte_sec + iofst);
  
//   uint32_t page_length = PageLength(0); // page size in words (4bytes)
  
//   int skip=page_length/srate; // ratio of acquisition-rate/output rate   
//   if(skip<1){
//     fprintf(stderr,"Error: Pagelength(%d)/srate(%d) <1 \n",
// 	    page_length, srate);
//     return  -1;
//   }
//   int ok_flag=0;
//   vector <uint8_t>::iterator ptr = arr.begin() + iofst;
//   int32_t ibyte(0);
//   string mode = vlbi_frame->DAS_Inf().Mode();

//   if(mode.find("/2bit/16ch") < mode.npos){
//     for( ibyte=0;ibyte<rq_Byte_sec && ptr < arr.end();ibyte+=4){
//       *ptr =enc_table[page_ptr[1*skip*(ibyte)+0]]; ptr++;
//       *ptr =enc_table[page_ptr[1*skip*(ibyte)+1]]; ptr++;
//       *ptr =enc_table[page_ptr[1*skip*(ibyte)+2]]; ptr++;
//       *ptr =enc_table[page_ptr[1*skip*(ibyte)+3]]; ptr++;
//     } ok_flag=1;
//   } 
//   if(mode.find("/2bit/08ch")< mode.npos){  
//     for(ibyte=0;ibyte<rq_Byte_sec && ptr < arr.end();ibyte+=4){
//       *ptr = enc_table[page_ptr[2*skip*(ibyte+0)+0+igrp*2]]; ptr++;
//       *ptr = enc_table[page_ptr[2*skip*(ibyte+0)+1+igrp*2]]; ptr++;
//       *ptr = enc_table[page_ptr[2*skip*(ibyte+2)+0+igrp*2]]; ptr++;
//       *ptr = enc_table[page_ptr[2*skip*(ibyte+2)+1+igrp*2]]; ptr++;
//     } ok_flag=1;
//   }
//   if(mode.find("/2bit/04ch")< mode.npos){
//     for(ibyte=0;ibyte<rq_Byte_sec && ptr < arr.end();ibyte+=4){
//       *ptr = enc_table[page_ptr[4*skip*(ibyte+0)+0+igrp]]; ptr++;
//       *ptr = enc_table[page_ptr[4*skip*(ibyte+1)+0+igrp]]; ptr++;
//       *ptr = enc_table[page_ptr[4*skip*(ibyte+2)+0+igrp]]; ptr++;
//       *ptr = enc_table[page_ptr[4*skip*(ibyte+3)+0+igrp]]; ptr++;
//     } ok_flag=1;
//   }
  
//   if(mode.find("/2bit/02ch")<mode.npos){
//     for(ibyte=0;ibyte<rq_Byte_sec && ptr < arr.end();ibyte+=4){
//       *ptr = enc_table[    (page_ptr[8*skip*(ibyte)+0]&0x0f)
// 		       |16*(page_ptr[8*skip*(ibyte)+4]&0x0f)]; ptr++;
//       *ptr = enc_table[    (page_ptr[8*skip*(ibyte)+0]&0x0f)
// 		       |16*(page_ptr[8*skip*(ibyte)+12]&0x0f)]; ptr++;
//       *ptr = enc_table[    (page_ptr[8*skip*(ibyte)+0]&0x0f)
// 		       |16*(page_ptr[8*skip*(ibyte)+20]&0x0f)];  ptr++;
//       *ptr = enc_table[    (page_ptr[8*skip*(ibyte)+0]&0x0f)
// 		       |16*(page_ptr[8*skip*(ibyte)+28]&0x0f)];  ptr++;
//     } ok_flag=1;
//   }

//   if(ok_flag==0){
//     cerr << "E K5_VSI::Convert(): mode(" << mode << ") is not supported\n";
//     Logsystem::error_logger()->Tstamp()
//       << "E K5_VSI::Convert(): mode(" << mode << ") is not supported\n";
//     return -1;
//   }
//   if( ibyte < rq_Byte_sec){
//     cerr << "E K5_VSI::Convert(): End of array before converting all data\n"
// 	 << "\tConverted =" << ibyte << "Byte, requested =" << rq_Byte_sec
// 	 << "Byte, arr.size=" << arr.size()  << " Bytes.\n";
//     Logsystem::error_logger()->Tstamp()
//       << "E K5_VSI::Convert(): End of array before converting all data\n"
//       << "\tConverted =" << ibyte << "Byte, requested =" << rq_Byte_sec
//       << "Byte, arr.size=" << arr.size()  << " Bytes.\n";
//     return -1;
//   }
//   return ibyte+iofst;
// }

const K5_VSI::IndevReturns_t  K5_VSI::UnMapping(const time_t epoch){
    if(munmap(page_ptr,4*PageLength(rpage)) !=0){
    	perror("Error in K5_VSI::UnMapping()");
      return GEN_ERROR;
    }
    ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF); // add 2014/10/27 miya
    usleep(INTVL_IOCTL_USEC); // add miya 2014/10/27
    return NORMAL;
}


int K5_VSI::FirstPage(){
  time_t tmin=0, twork;
  //  k5vsi_mutex.lock();
  for(int i=0; i<npage; i++){
    //    if( (twork = ioctl(device,GET_FRAME_EPOCH,i)) > 0){
    if(Logsystem::Test_Logsys(Logsystem::DEBUG, 2)) 
      Logsystem::debug_logger()->Tstamp()
	<< "D K5_VSI::FirstPage(): Calling GetFrameEpoch(" << i << ")\n";
    if( (twork = GetFrameEpoch( i ) ) > 0){
      if(tmin == 0 && twork > 0) tmin = twork;
      if(tmin > twork) tmin = twork;
    }else{
      cerr << "E K5_VSI::FirstPage():Failed to get epoch(" << i << ")" 
	   << strerror(errno) << endl;
      Logsystem::error_logger()->Tstamp()
	<< "E K5_VSI::FirstPage():Failed to get epoch(" << i << ")" 
	<< strerror(errno) << "\n";
      
    }
  }
  //  k5vsi_mutex.unlock();

  //  int ipg = (tmin + npage/2) % npage; ?? by M.S.2012.1.5
  int ipg = tmin  % npage;
  return ipg;
}

// const K5_VSI::IndevReturns_t 
// K5_VSI::MappingData(vector <uint8_t>& arr,time_t epoch){
//   cerr << "Error @MappingData(): This is dummy function in K5_VSI."
//        <<" Use other MappingData() instead." << endl;
//   return GEN_ERROR;
// }; 



