/** 
 * @Version $Rev: 181 $
 * @Author M.Sekido
 * @Date  $Date: 2016-02-05 16:50:58 +0900 (Fri, 05 Feb 2016) $
 * $Id: Rate_meas.cpp 181 2016-02-05 07:50:58Z vlbi $ 
 */

#include "Rate_meas.hpp"
#include <algorithm>
using namespace std;

/// Constructor 
Rate_Meas::Rate_Meas():
  rcv_Mbps(0),snd_Mbps(0),rcv_total(0),snd_total(0),
  rcv_pld_total(0),snd_pld_total(0),
  run_state(false),sock_(0),update_intvl(DEFAULT_UPDATETIME),
  disp_sw(Rate_Meas::OFF),target_rate_Mbps(0), workmode(TX),
  pktRegTbl(DEFAULT_NPKT),cntMis(0),seq0(-1),_acksock(0),ackp(0),AckSend_ONOFFstate(false),Ack_ONOFFstate(false),Ack_state(true),ack_data_port(ACK_DEFAULT_DATA_PORT) // '_acksock(0),ackp(0)' add 2015/4/27 miya 'AckSend_ONOFFState(false)Ack_state(false)Ack_ONOFFstate(false)' add 2015/6/23 'ack_data_port(ACK_DEFAULT_DATA_PORT)' add 2016/2/3
{ 
  thread_number++;
  pktCnt[0] =pktCnt[1] = 0;
  cntDup[0] = cntDup[1] = 0;
  iseq_max[0] = iseq_max[1] = 0;
  RepktCnt = 0; //add 2015/4/27 miya
  ReInit();
  char buf[256];
  sprintf(buf,"RM%d", thread_number);
  SetID(buf); // Set Thread ID
  memset(&ack_addr,0,sizeof(ack_addr));  // add 2015/4/27 miya 
}


void *Rate_Meas::thread_func(void){
  gettimeofday(&st,NULL);  
  st0 = st;
  run_state=true;
  Measurement();
  return NULL;
} 


void Rate_Meas::stop(){
  //  usleep(500000);
  if(run_state) {
    run_state=false; 
    //    if(join() ==0){
    //    Thread::stop();
    Logsystem::debug_logger()->Tstamp()
      << "Rate_Meas[" << Thread::StrID()
      << "]::Measuremet() stop-joined.\n";
    Logsystem::debug_logger()->Flush();
    cerr << "Rate_Meas[" << Thread::StrID() 
	 << "]::Measuremet() stop-joined." << endl;
    //    }
  }
  ReInit();
}


string Rate_Meas::Epoch()const{
  struct tm tm;
  string t;
  if(gmtime_r(&et.tv_sec,&tm) ==NULL){
    cerr << "Error in Rate_Meas::Epoch():Failed gettime_r()\n";
    return t;
  }
  char buf[80];
  strftime(buf, 80,"%Y/%j %H:%M:%S", &tm);
  t = buf;
  return t;
}

void Rate_Meas::ReInit(const int flag){
  if(flag <1){
    rcv_Mbps = snd_Mbps = 0.0;
    pld_rcv_Mbps = pld_snd_Mbps = 0.0;
    rcv_total = snd_total = rcv_total_st0 = snd_total_st0 = 0;
    rcv_pld_total = snd_pld_total = rcv_pld_total_st0 = snd_pld_total_st0 = 0;
    recomended_wait = 0.0;
    seq0 = -1; 
  }
  cntDup[0] = 0;
  pktCnt[0] = 0;
  iseq_max[0] = 0;
  RepktCnt = 0; //add 2015/4/27 miya
  fill(pktRegTbl.begin(),pktRegTbl.end(),false); 
}


void Rate_Meas::TakeStatistics(){
  //  cerr << "Computing Statistics" << endl;
  _mutex.lock();
  cntMis = count(pktRegTbl.begin(),pktRegTbl.begin() + iseq_max[0], false);
  cntDup[1] = cntDup[0]; // Rsearve last counter value;
  if(pktCnt[0] != 0 && 
     pktCnt[0] != iseq_max[0]+1){ // Because iseq start from 0;
    Logsystem::debug_logger()->Tstamp()
      << "Warning pktCnt(" << pktCnt[0] << ")!= iseq_max("
      << iseq_max[0] << ")\n";
    Logsystem::debug_logger()->Flush();
    
  }
  pktCnt[1] = pktCnt[0]; // Store the counter data to register 
  iseq_max[1] = iseq_max[0];

  ReInit(1);
  seq0=-1;
  _mutex.unlock();

}
// add 2015/4/27 miya start----------------------------------
void Rate_Meas::Set_Ack(){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  vtp_ack.ack_epoch = (uint32_t)(tv.tv_sec);
  vtp_ack.ack_nano_Epoch = (uint32_t)(tv.tv_sec*1000);
  vtp_ack.ack_seq = (uint64_t)iseq_max[1];
  vtp_ack.ack_recv = (uint64_t)pktCnt[1]-(uint64_t)cntDup[1];
  vtp_ack.ack_order = (uint64_t)RepktCnt;
}

int Rate_Meas::Create_Acksock(){
	if(_acksock!=0) delete _acksock;
	_acksock = new UDPSocket(ack_pktsize);
	_acksock->set_so_sndbuf(ACK_SNDBUFSIZE);
    return 0;
}

bool Rate_Meas::Ack_Ready(const int ack_port_no){
	
  uint16_t port;
	
  Create_Acksock();
	
  if(ack_port_no ==0){
  	port = ack_data_port;
  }else{
    port = ack_port_no;}

	if ( AckSend_ONOFFState() ){ // add miya 2015/6/23
    	fprintf(stderr,"Set_AckHost(%s) Set_AckPort(%d) \n", AckHost().c_str(),port);
	    struct hostent *hostinfo =  gethostbyname(AckHost().c_str() );
	    if( hostinfo == NULL){
	      Logsystem::error_logger()->Tstamp() 
		<< "E Rate_meas::Ack_Ready(): Could not get hostinfo.\n";
	      cerr << "Error in Rate_meas::Read(): Could not get hostinfo."
		   << endl;
	      return false;
	    }
	    ack_addr.sin_port=htons((uint16_t)port);
	    ack_addr.sin_family=AF_INET;
		ack_addr.sin_addr.s_addr = //inet_addr("127.0.0.1");
		*((unsigned long *)hostinfo->h_addr_list[0]);
	    char ip_str[80];    
	    inet_ntop(AF_INET, &ack_addr.sin_addr, ip_str, sizeof(ip_str));
	    fprintf(stderr,"Set_AckHost(%s): Destination IP=%s:%d mtu=%d\n", 
		    ack_hostname.c_str(),ip_str, port, _acksock->get_mtu());
		return true;
	}else{
	  string ip;
	  int ret = _acksock->bind(ip, port); 
	  if(ret < 0){
	    cerr << "E Rate_meas::Ack_Ready(): Failed to get address from Socket:"
		 << strerror(errno) << endl;
	    return false;
	  }		
	    ack_addr.sin_family=AF_INET;
	    ack_addr.sin_port=htons(port);
	    ack_addr.sin_addr.s_addr = INADDR_ANY ;
	}
	return true;
}

void Rate_Meas::Ack_Close(){
	_acksock->close();
}
// add 2015/4/27 miya end----------------------------------
