
#ifndef __RATE_MEAS__
#define __RATE_MEAS__
/** Rate Measurement Class Header 
 * @Version $Rev: 181 $
 * @Author M.Sekido
 * @Date  $Date: 2016-02-05 16:50:58 +0900 (Fri, 05 Feb 2016) $
 * $Id: Rate_meas.hpp 181 2016-02-05 07:50:58Z vlbi $ 
 */

#include "UDPSocket.h" //add 2015/4/27 miya
#include "netutl.h" // add miya 2015/6/23

#include <string>
#include <vector>
#include <iostream>
#include <cmath>
using namespace std;

#include <Socket.h>
#include <Logsystem.hpp>
#include <Thread.h>

extern "C"{
#include <unistd.h>
#include <arpa/inet.h>  // add miya 2015/6/23
#include <netinet/in.h>  // add miya 2015/6/23
#include <sys/socket.h>  // add miya 2015/6/23
}

const double DEFAULT_UPDATETIME = 1.0;
const int    DEFAULT_NPKT       = 300000; // miya 200000->300000
const int ERROR_ACTCNT_LIMIT  =  10; // add miya 2015/6/23

// add 2015/4/27 miya start----------------------------------
typedef struct{
	uint32_t ack_epoch;
	uint32_t ack_nano_Epoch;
	uint64_t ack_seq;
	uint64_t ack_recv;
	uint64_t ack_order;
} Ack_Packet;
const int ACK_DEFAULT_DATA_PORT  = 29000;
const int ACK_DEFAULT_BUF_SIZE  = 1500;
const int ACK_SNDBUFSIZE    =  1500;
const int ack_pktsize = 32; 
// add 2015/4/27 miya end----------------------------------

class Rate_Meas:public Thread {
public:
  /// Running mode 
  enum Mode_t {TX, RX, BOTH};
private:
  static int thread_number; ///< Thread number 
  enum Disp_sw_t {RX_ON, TX_ON, OFF};
  struct timeval  st0;  ///< Overall Initial epoch 
  struct timeval  st;   ///< Initial Epoch
  struct timeval  et;   ///< Final Epoch
  Mutex  _mutex;        ///< mutex to access-control for UDP counter
  double rcv_Mbps ;  
  double snd_Mbps ;  
  double pld_rcv_Mbps ;  ///< Receive Data rate of contents.
  double pld_snd_Mbps ;  ///< Sending Data rate of contents.
  uint64_t rcv_total;  ///< Total Bytes of packet (UDP)
  uint64_t snd_total;  ///< Total Bytes of packet 
  uint64_t rcv_total_st0;  ///< Total Bytes of packet (UDP)
  uint64_t snd_total_st0;  ///< Total Bytes of packet 
  uint64_t rcv_pld_total; ///< Total Receive Net Payload excluding Frame Header 
  uint64_t snd_pld_total; ///< Total Sent Net Payload excluding Frame Header 
  uint64_t rcv_pld_total_st0; ///< Total Receive Net Payload excluding Frame Header 
  uint64_t snd_pld_total_st0; ///< Total Sent Net Payload excluding Frame Header 
  bool run_state; 
  Socket *sock_;
  double update_intvl; ///< Update Interval
  Disp_sw_t disp_sw;       ///< Display Switch 1: Receiving, 2: Sending, 0: Off
  double recomended_wait;  ///< Recomended waiting time for target rate
  double target_rate_Mbps;   ///< Target rate in Mbps
  Mode_t workmode;  ///< Working mode RX or TX 
  vector <bool>  pktRegTbl; ///< Packet seq. Number register 
  int    cntDup[2];    ///< Duplicated Packet counter, current[0] and last[1]
  int    cntMis;    ///< Missing Packet counter
  int32_t pktCnt[2];    ///< Packet count number /sec counter and regiter
  int64_t seq0;      //< Initial sequential number for counting
  int32_t iseq_max[2];   //< Maximum sequential number for one sec. current and reg.
  int32_t RepktCnt;    //add 2015/4/27 miya
  /** Initialize state parameters 
   *  @param flag:
   *      - 0: Initialize all paramters
   *      - 1: Initialize only UDP parametes(mis,dup,pktRegTbl[])
   */
  void ReInit(const int flag=0);
  Socket *_acksock;  ///< Socket for data transfer //add 2015/4/27 miya
  /** socket address structure used for UDP */
  struct sockaddr_in ack_addr; //add 2015/4/27 miya
  int Create_Acksock();  ///< Creating Scoket add 2015/4/27 miya
  bool Ack_Ready(const int ack_port_no=0);  //add 2015/4/27 miya
  void Ack_Close(); ///< Closing connections add 2015/4/27 miya
  void Set_Ack();  ///  add 2015/4/27 miya
  Ack_Packet vtp_ack; // add 2015/4/27 miya
  int ackp; // add 2015/4/27 miya
  bool AckSend_ONOFFstate; // add miya 2015/6/23
  bool Ack_ONOFFstate; // add miya 2015/6/23
  bool Ack_state; // add miya 2015/6/23
  vector <uint8_t> packet; // Packet add miya 2015/6/23
  string ack_hostname; //  add miya 2015/6/23
  int addr_len; // add miya 2015/6/23
  int mtu;// add miya 2015/6/23
  int cnt_ack; // add miya 2015/6/23
  int ack_data_port;  // add miya 2016/2/3 < Ack Data port
public:
  Rate_Meas();   ///< Constructor 
  ~Rate_Meas(){stop(); thread_number--;} ///< Destructor 
  void stop();   ///< Stop measurement 
  void* thread_func(void); ///< Thread function
  /// Set Socket to make measurement
  int Set_socket(Socket* const so){if(so==0) return -1; sock_ = so; return 0;}
  int Measurement();  ///< Main function for rate measrement
  /// Returns Rx data rate in Mbps
  double Rx_Mbps()const {return rcv_Mbps;}  ///< Returns Rx data rate
  /// Returns Tx data rate in Mbps
  double Tx_Mbps()const {return snd_Mbps;}  ///< Returns Tx date rate
  /**  Returns Payload Rx data rate excluding frame header */
  double Rx_Pld_Mbps()const {return pld_rcv_Mbps;}  
  /**  Returns Payload Tx date rate excluding frame header */
  double Tx_Pld_Mbps()const {return pld_snd_Mbps;}  
  string Epoch()const;  ///< Returns Epoch in string 
  double T_Elapsed();   ///< Returns Elapsed time (socket life time)
  /// Returns Total Received bytes
  uint64_t Total_rcv()const{ return rcv_total;}
  /// Returns Total Sent bytes
  uint64_t Total_snd()const{ return snd_total;}
  //  void Disp_sw(Disp_sw_t st=OFF){disp_sw=sw;}  /// Display Switch
  /// Set Displaying Rx data rate
  void Disp_Rxon(){disp_sw=RX_ON;}
  /// Set Displaying Tx data rate
  void Disp_Txon(){disp_sw=TX_ON;}
  /// Set Displaying off of measured data
  void Disp_Off(){disp_sw=OFF;}
  /// Return running or not
  bool is_running()const{ return run_state;}
  /// Register Receved Payload for counting net data rate.
  void Add_Rcv_Pld(const uint32_t byte, const int64_t sq=-1){
    rcv_pld_total += byte; if(sq != -1) Add_Cnt(sq);}
  /// Register Sent Payload for counting net data rate.
  void Add_Snd_Pld(const uint32_t byte){ snd_pld_total += byte;}
  /// Set Target rate (Mbps)
  void Set_Target_rate(const double Mbps){target_rate_Mbps = Mbps;}
  /// Returns Recomended wait time to achieve target rate.
  double Rcmnd_wait() const {return recomended_wait;}
  /// Set Update frequency of rate measurement.
  void SetUpdateFreq(const double Hz){update_intvl = 1./Hz;}
  /// Returns target Tx data rate (Mbps)
  double TargetRate()const{return target_rate_Mbps;} // (Mbps)
  /// Set Measuermet mode (RX/TX/Both)
  void Set_Mode(const Mode_t m){workmode =m;};
  /// Register the sequential number for UDP (mis/dup) statistics 
  void Add_Cnt(const int64_t seq);
  /// Computing statistics of received traffic
  void TakeStatistics();
  /// Returns Packet count until previous update
  int32_t PktCnt() const {return pktCnt[1];}
  const int Nthread() const {return thread_number;}
  void Set_AckSend_state(const bool AckSend){AckSend_ONOFFstate = AckSend;}// add miya 2015/6/23
  bool AckSend_ONOFFState()const{ return AckSend_ONOFFstate;} // add miya 2015/6/23
  void Set_Ack_state(const bool Ack){Ack_ONOFFstate = Ack;}// add miya 2015/6/23
  bool Ack_ONOFFState()const{ return Ack_ONOFFstate;} // add miya 2015/6/23
  bool Ack_State()const{ return Ack_state;} // add miya 2015/6/23
  void Set_AckHost(const char* h){ack_hostname=h;} // add miya 2015/6/23
  string AckHost() {return ack_hostname;} // add miya 2015/6/23
  void Set_AckDataPort(const int p) {ack_data_port=p;} // add miya 2016/2/3
};

/** Add_Cnt() 
 *  @doc Add Counting for SUDP packet statistics
 *      counting up cntDup, cntMis counter 
 */
inline void Rate_Meas::Add_Cnt(const int64_t seq){
  _mutex.lock();
  pktCnt[0]++;
  if( seq < 0){
    Logsystem::warning_logger()->Tstamp() 
      << "Warning at Rate_Meas::SUDP_Cnt():Sequential number is negative("
      << seq << ")\n" ;;
    return ;
  }
  if(seq0 == -1) seq0 = seq;
  
  int32_t iseq = seq - seq0;
  if( iseq < iseq_max[0] ) { RepktCnt++; }//add 2015/4/27 miya
  if( iseq > iseq_max[0] ) {iseq_max[0] = iseq;}
  
  if( (int64_t)pktRegTbl.size()+1 < iseq ){
    Logsystem::warning_logger()->Tstamp() 
      << "Warning at Rate_Meas::SUDP_Cnt():pkt number("
      << iseq << ") is larger than regTable.size(" << (int)pktRegTbl.size() 
      << ")\n" ;
    pktRegTbl.resize(iseq);
  }

  if(pktRegTbl[iseq])cntDup[0]++;
  else  pktRegTbl[iseq] = true;
  _mutex.unlock();
}

inline double Rate_Meas::T_Elapsed(){
  double ret= (double)(et.tv_sec - st0.tv_sec 
		       + (et.tv_usec - st0.tv_usec)*1.e-6);  
  return ret;
}

inline int Rate_Meas::Measurement(){
  Logsystem::debug_logger()->Tstamp() << "Rate_Meas::Measurement() started\n";
  if(sock_ ==0){
    cerr << "Error in Rate_Meas::Measurement(): socket is not set\n";
    stop();
    return -1;
  }

// add 2015/4/27 miya start----------------------------------
	cnt_ack=0; // add miya 2015/6/23
	if ( AckSend_ONOFFState() ){ // add miya 2015/6/23
	  if( ! Ack_Ready() ){
	    Logsystem::error_logger()->Tstamp()
	      << "E Rate_Meas::Ack_Ready(): failed.\n";
	  }
	}
	else if ( Ack_ONOFFState() ){
	  if( ! Ack_Ready() ){
	    Logsystem::error_logger()->Tstamp()
	      << "E Rate_Meas::Ack_Ready(): failed.\n";
	  	}
		addr_len = sizeof(ack_addr); // add miya 2015/6/23
	  	mtu = _acksock->get_mtu();// add miya 2015/6/23
	  	packet.resize(mtu);// add miya 2015/6/23
		_acksock->Set_Recv_flg(1);// add miya 2015/7/27
	}
// add 2015/4/27 miya end----------------------------------
	
  //  struct timeval temp;  
  //  double t1,t2,delta, delta0;
  //  uint32_t mdif, mdif_pld;

  //  long  dnsec;   long  dsec;
  rcv_total = sock_->get_bytes_rcvd();
  snd_total = sock_->get_bytes_sent();
  uint64_t last_rcv  =rcv_total;
  uint64_t last_snd = snd_total;
  uint64_t last_rcv_pld = rcv_pld_total;
  uint64_t last_snd_pld = snd_pld_total;
  //  uint64_t curnt_snd_pld(0), curnt_rcv_pld(0); 

  gettimeofday(&st0, NULL);
  //    double t0 = st0.tv_sec + st0.tv_usec*1.e-6;
  //  uint64_t rcv_total_st0 = rcv_total;
  //  uint64_t snd_total_st0 = snd_total;
  //  uint64_t rcv_pld_st0 = rcv_pld_total;
  //  uint64_t snd_pld_st0 = snd_pld_total;

  gettimeofday(&et, NULL);
  struct timespec req;
  req.tv_sec =0;
  req.tv_nsec = (long)(1000000L - et.tv_usec);
  nanosleep(&req,0);

  int cnt=0;
  uint32_t mdif, mdif_pld;
  // mdif_st0, mdif_pld, mdif_pld_st0;
  int32_t update_usec_hf = (int32_t)(update_intvl*500000); // micro sec
  
  while(run_state){
    rcv_total = sock_->get_bytes_rcvd();
    snd_total = sock_->get_bytes_sent();
    uint64_t curnt_rcv_pld = rcv_pld_total;
    uint64_t curnt_snd_pld = snd_pld_total;
    gettimeofday(&et, NULL);
    //    double t1 = st.tv_sec + st.tv_usec*1.e-6;
    //    double t2 = et.tv_sec + et.tv_usec*1.e-6;
    //    double delta = t2 - t1;
    double delta = (et.tv_sec - st.tv_sec) + (et.tv_usec - st.tv_usec)*1.e-6;
    //	double delta0 = t2 - t0;
    //    Logsystem::debug_logger()->Tstamp() << "Rcv Byte:" << rcv_total 
    //					<< " Snd Byte:" << snd_total << endl;

    if(workmode != TX){
      /** Receiving data rate */
      mdif = rcv_total - last_rcv;
      //      mdif_st0 = rcv_total - rcv_total_st0;
      mdif_pld = curnt_rcv_pld - last_rcv_pld;
      //      mdif_pld_st0 = curnt_rcv_pld - rcv_pld_st0;
      rcv_Mbps = (double)mdif/delta * 8.e-6;  // Mbps
      //    pld_rcv_Mbps = (double)mdif_pld/delta*8.e-6;  // Mbps
      pld_rcv_Mbps = (double)mdif_pld/delta * 8.e-6;  // Mbps
    }
    
    if(workmode != RX){
      /** Sending data rate */
      mdif = snd_total - last_snd;
      //      mdif_st0 = snd_total - snd_total_st0;
      mdif_pld = curnt_snd_pld - last_snd_pld;
      //      mdif_pld_st0 = curnt_snd_pld - snd_pld_st0;
    
      snd_Mbps = (double)mdif/delta * 8.e-6;  // Mbps
      // snd_Mbps = (double)mdif_st0/delta0*8.e-6;  // Mbps
      pld_snd_Mbps = (double)mdif_pld/delta*8.e-6;  // Mbps
      //pld_snd_Mbps = (double)mdif_pld_st0/delta0*8.e-6;  // Mbps
      // 	    Logsystem::debug_logger()->Tstamp()
      // 		<< "snd_total(" << snd_total << ")-last_snd(" << last_snd
      // 		<< ")=mdif" << mdif << ", currnt_snd_pld(" << curnt_snd_pld
      // 		<< ")-last_snd_pld(" << last_snd_pld << ")=mdif_pld(" 
      // 		<< mdif_pld << endl;
    }


    double dt0 = T_Elapsed(); // sec. from st0
    float avg_rate(0);
    int cnt_sec=0;
    double RatioDup(0.0), RatioMis(0.0);
    string msg;
      
    /** Display work within update_interval*0.5 from 1 sec tick */
    if(et.tv_usec < update_usec_hf || et.tv_usec > 1000000 - update_usec_hf ){
      //    if(1){
      char buf[2048], buf2[2048];
      switch(disp_sw){
      case RX_ON: // Rx rate measure
	TakeStatistics();
	sprintf(buf,"[%s] RxRate[%s]:%6.2lf (net:%6.2lf) Mbps",  
		Epoch().c_str(), Thread::StrID().c_str(), 
		rcv_Mbps, pld_rcv_Mbps);
	// 		fprintf(stderr,"[%s] RxRate:%6.2lf (net:%6.2lf) Mbps",  
	// 			Epoch().c_str(), rcv_Mbps, pld_rcv_Mbps);
	msg = buf;
	avg_rate = (float)(rcv_total*8.*1.e-6/dt0);
	RatioDup = iseq_max[1] == 0 ? 0.0 
	  : (double)cntDup[1]/(iseq_max[1]+1)*100.0;
	RatioMis = iseq_max[1] == 0 ? 0.0
	  : (double)cntMis/(iseq_max[1]+1)*100.0;
	sprintf(buf2,
		" Total(%.2lf MByte)/%.2lf sec = %6.2lf Mbps Dup=%4d(%4.1lf%%) Mis=%4d(%4.1lf%%)\n",
		1.e-6*rcv_total, dt0, avg_rate, cntDup[1], RatioDup, cntMis, RatioMis);
	msg += buf2;
	cerr << msg;
	*Logsystem::info_logger()  << msg;
	// 		fprintf(stderr,
	// 		" Total(%.2lf MByte)/%.2lf sec = %6.2lf Mbps Dup=%4d(%4.1lf%%) Mis=%4d(%4.1lf%%)\n",
	// 			1.e-6*rcv_total, dt0, avg_rate, cntDup[1], RatioDup, 
	// 			cntMis, RatioMis);
	break;
      case TX_ON: // Tx rate measure
	sprintf(buf,"[%s] TxRate[%s]:%6.2lf (net:%6.2lf) Mbps",  
		Epoch().c_str(), Thread::StrID().c_str(), 
		snd_Mbps, pld_snd_Mbps);
	// 		fprintf(stderr,"[%s] TxRate:%6.2lf (net:%6.2lf) Mbps",  
	// 			Epoch().c_str(), snd_Mbps, pld_snd_Mbps);
	msg = buf;
	avg_rate = (float)(snd_total*8.*1.e-6/dt0);
	sprintf(buf2," Total(%.2lf MByte)/%.2lf sec = %6.2lf Mbps\n",
		1.e-6*snd_total, dt0, avg_rate);
	msg += buf2;
	cerr << msg;
	*Logsystem::info_logger() << msg ;
	Logsystem::info_logger()->Flush();
	// 		fprintf(stderr," Total(%.2lf MByte)/%.2lf sec = %6.2lf Mbps\n",
	// 			1.e-6*snd_total, dt0, avg_rate);
	break;
      case OFF: // Display Off;
	break;
      }
      // add 2015/4/27 miya start----------------------------------
      if ( AckSend_ONOFFState() ){ // add miya 2015/6/23
	    Set_Ack();
		ackp = _acksock->sendto((struct sockaddr*)&ack_addr, sizeof(ack_addr), (char*)&vtp_ack, ack_pktsize);
	    if( ackp < 0 ){
	      cerr << "Warinig: Rate_Meas::Measurement(): Ack_packets\n";
	    }
      }
      // add 2015/4/27 miya end----------------------------------
      // add 2015/6/23 miya start----------------------------------
      if ( Ack_ONOFFState() ){ 
      	int nread = _acksock->recvfrom((struct sockaddr&)ack_addr, addr_len, (char*)&packet[0], mtu);
	    if( nread < 0){
	      cerr << "E Rate_Meas::Measurement(): Ack_packets_recv:" 
		   << strerror(errno) << endl;
	    	cnt_ack++;
	    	if( cnt_ack > ERROR_ACTCNT_LIMIT){
	    		Ack_state = false;
	    	}
	    }else{ cnt_ack=0; }
      	
      }// add 2015/6/23 miya end----------------------------------
      //      gettimeofday(&st, NULL); // Every second start time
      cnt_sec=0;
	    
      last_rcv_pld = curnt_rcv_pld;
      last_snd_pld = curnt_snd_pld;
      last_rcv = rcv_total;
      last_snd = snd_total;
      st=et;  	// Set Current time to next start time.    
    }

    if(workmode != RX){
      if(cnt== 0){
		Logsystem::warning_logger()->Tstamp() 
		  << "Warinig: Rate_Meas::Measurement(): cnt=0,Skip recomended_wait set.\n";
      }else{  
		if( target_rate_Mbps == 0){
		  Logsystem::warning_logger()->Tstamp() 
		    << "Warinig: Rate_Meas::Measurement(): Target rate is not set\n";
		}else{
		  recomended_wait = snd_pld_total*8.e-6/target_rate_Mbps - dt0;
	// 	  if(Logsystem::debug_logger()->Level() >= Logger::DETAIL ){
	// 	    Logsystem::debug_logger()->Tstamp() << "snd_pld_total(" <<
	// 	      snd_pld_total*8.e-6 << ")/target(" <<target_rate_Mbps << ") -t0(" 
	// 						<< dt0 << ")\n";
	// 	  }
			}
      }
    }



    /** Sleep for a interval */
    struct timeval temp;
    gettimeofday(&temp,NULL);
    double t1 = temp.tv_sec + temp.tv_usec*1.e-6;
    delta = (double)st.tv_sec + cnt_sec * update_intvl;
    while(delta <= t1 ){
      cnt_sec ++;
      delta += update_intvl;
    }
    delta -= t1;
    //	cerr << "Sleep = delta(" << delta << ")" << endl;
    if(delta < 0) continue;

//     req.tv_sec =(time_t)floor(delta);
//     req.tv_nsec = (long)((delta - req.tv_sec)*1000000000);
//     //    if(nanosleep(&req, NULL) ==-1){
//     while(nanosleep(&req, &req) ==-1){
//       perror("Error in Rate_Meas::start(): failed nanosleep()"); 
//       //      return -1;
//     } 
    
    useconds_t ut = (useconds_t)(delta * 1000000);
    usleep(ut);
	
    cnt++;
  }
    
  switch(workmode){
  case RX:
    *Logsystem::info_logger() << "I Exiting RXRate_Meas[" << Thread::StrID()
			      << "]::Measurement().\n";
    break;
  case TX:
    *Logsystem::info_logger() << "I Exiting TXRate_Meas[" << Thread::StrID()
			      << "]::Measurement().\n";
    break;
  default:
    *Logsystem::info_logger() << "I Exiting RxTxRate_Meas[" << Thread::StrID()
			      << "]::Measurement().\n";
    break;
  }
  return 0;
}

#endif
