/** 
 * @Version $Rev: 181 $
 * @Author M.Sekido
 * @Date  $Date: 2016-02-05 16:50:58 +0900 (Fri, 05 Feb 2016) $
 * $Id: Trans_ctl.cpp 181 2016-02-05 07:50:58Z vlbi $ 
 */

#include "Trans_ctl.hpp"
using namespace std;

// Constructor 
TransCtl::TransCtl():
  //  sock_(0),ctrl_state(IDLE),out_protocol(0),indev(0),ngrp(0),
  sock_(0),ctrl_state(IDLE),indev(0),ngrp(0),
  nout(0),nin(0),ctrl_port(DEFAULT_CTRL_PORT) 
{  
  out_protocol.resize(0);
  sock_ = new TCPSocket(DEFAULT_MTU_SIZE);
  dumper.resize(1); // modify 2015/9/14 1->0 miya 2015/10/16 0->1
  ///dumper[0] = new Dumper;}  // comment 2015/3/9 miya
}

void TransCtl::Clean() // Cleanup for Destructor 
{  ;
  if(sock_!=0){ 
    delete sock_; 
    for(int i=0; i< (int)dumper.size(); i++) delete dumper[i];
    dumper.resize(0);
    sock_=0;
  }
  //  indev->join();
}


// TransCtl::~TransCtl() // Destructor 
// {  ;
//   if(sock_!=0){ 
//     delete sock_; 
//     for(int i=0; i< (int)dumper.size(); i++) delete dumper[i];
//     dumper.resize(0);
//     sock_=0;
//   }
//   //  indev->join();
// }



/** Reuesting Transmission to server  */
void* TransCtl::thread_func()
{
  ctrl_state = SERVING; //  server running
  //  int err = 0;
  
  if(Listen() <0){
    Logsystem::error_logger()->Tstamp()
      << "Failed to setup socket for listen.\n";
    Logsystem::error_logger()->Tstamp()  << " Exiting TrancCtrl thread.\n";
    ctrl_state = TERM;
    //    err ++;
  }else{
    //    if(indev->Ready()){
    while(ctrl_state != TERM ){
      if(CommunicateWithClient() <0){
	string msg = "TransCtrl::CommunicateWihtClient() ending.\n";
	Logsystem::info_logger()->Tstamp() << msg;
	cerr << msg;
      }
      fprintf(stderr, "I TransCtl::thread_func): state=%s\n",
	      StrCtrlState().c_str());

    }
    //    }
    ctrl_state = IDLE;
  }
  
  //  Thread::state = Thread::ENDING;
  if( indev->Status() == Thread::RUNNING ) indev->cancel();
  return 0;
}


int TransCtl::Set_ApplProto(VLBI_Proto* vp[], const int ng){
  out_protocol.resize(ng);
  ngrp = ng; // Number of Groups
  for(int i=0; i<ng; i++) out_protocol[i] = vp[i];
  //  out_protocol = vp;

  //  out_subgrp.resize(ngrp);
  //  in_subgrp.resize(ngrp);
  if(out_protocol[0]->ApplProtocol()==TransInfo::RAW_Proto){ // add 2014/12/9 miya 出力がfileの場合
    dumper[0] = new FileDumper; // add 2014/12/9 miya modify 2015/3/6 
    dumper[0]->Set_Nout(ngrp); // add 2014/12/9 miya
  }else{ // add 2014/12/9 miya 出力がNETの場合
    if( (int)dumper.size() <= ngrp ){ // modify 2015/10/16 miya  '<' -> '<='
      //int idmp = dumper.size();  comment 2015/9/14 miya
      dumper.resize(ngrp);
      for(int i=0; i<ngrp; i++)  dumper[i] = new Dumper;  //modify 2015/9/14 idmp->0 miya
    }
  }
  return 0;
}


int TransCtl::CommunicateWithClient(){

  fprintf(stderr,"Waiting for connection from Clnt...");
  int   ThreadSock = sock_->accept();
  if( ThreadSock  == -1){
    fprintf(stderr,"E CommunicateWithClient(): Failed Accept (errno=%d) \n", errno);
    perror("Accept Error");
    sock_->Error();
    return -1;
  }else{
    fprintf(stderr, "accepted\n");
  }
  
  TCPSocket*  tcp_sock = new TCPSocket(ThreadSock,sock_->get_mtu());
  tcp_sock->Print();
  uint8_t readbuf[256];
  while(ctrl_state == SERVING){
    try{
      //  int status = tcp_sock->recv((uint8_t *)&readbuf, sizeof(readbuf));
      int status = tcp_sock->readline(readbuf, sizeof(readbuf));
      fprintf(stderr,"Received msg: %s\n", readbuf);
      if ( status < 1) 	throw status;
      if(CmdParser((char*)readbuf,tcp_sock) <0) throw readbuf;
    }

    catch (const char* str){
      fprintf(stderr,"E CmdParser(): Invalid Command: %s \n", str);
      // return -1;
      continue;
    }
    catch (const unsigned char* str){
      fprintf(stderr,"E CmdParser(): Invalid Command: %s \n", str);
      continue;
      // return -1;
    }

    catch(int stat1){
      switch(stat1){
      case 0:
	//	fprintf(stderr,"Connection is closed by Client.\n");
	Logsystem::info_logger()->Tstamp()
	  << "I CommunicateWithClient(): Connection is closed by Client.\n";
	delete tcp_sock;
	return -1; 	break;
      case  -1:
	{
	  char buf[256];
	  sprintf(buf,"E CommunicateWithClient(): (errno=%d):%s \n", 
		  errno, strerror(errno));
	  Logsystem::error_logger()->Tstamp() << buf;
	  delete tcp_sock;
	  return -1;	break;
	}
      default:
	{
	  char buf[256];
	  sprintf(buf,"E CommunicateWithClient(): (errno=%d):%s \n", 
		  errno, strerror(errno));
	  Logsystem::error_logger()->Tstamp() << buf;
	  delete tcp_sock;
	  return -2;	break;
	}
      }
    } // Catch
    
  } // while
  
  fprintf(stderr, "Exiting TransCtl::CommunicateWithClient(): state=%s\n",
	  StrCtrlState().c_str());
  delete tcp_sock;
  Logsystem::info_logger()->disable_logging();
  Logsystem::error_logger()->disable_logging();
  return 0;

}

inline int TransCtl::Listen(){
  
  string thisip = "0.0.0.0";
  if( sock_->bind(thisip, ctrl_port) == -1){
    fprintf(stderr,"E  TransCtl::Listen(): failed to listed at (%s:%d) %s\n",
	    thisip.c_str(), ctrl_port, strerror(errno));
    //    perror("E  TransCtl::Listen()");
    return -1;
  }

  fprintf(stderr,"socket setup for Listening ...");
  if( sock_->listen() == -1){
    fprintf(stderr,"E in TransCtl::Listen(): failed to listen().\n");
    return -1;
  }else{
    fprintf(stderr," Successed\n");
    fprintf(stderr," Waiting for connection with port(%d)\n", ctrl_port);
  }

  return 0;
}


const int TransCtl::CmdParser(const char* cmd, Socket *sock){
  string str = cmd;
  string response;
  int ret(0);
  string::size_type pos;
  char msg[1024];
  
  if( (pos = str.find( "play_rate=data:" )) != str.npos ){
    string rate = str.substr( pos+15, str.npos-15 );
    int32_t Mbps = strtol( rate.c_str(),NULL,10 );
    if( Mbps > 0 && Mbps < 10000){ 
      for(int i=0; i<ngrp; i++)
	out_protocol[i]->Set_Speed(1000000/32*Mbps, true);
      //	out_protocol[i].Set_Speed(1000000/32*Mbps, true);
      char buf[256]; sprintf(buf,"!play_rate=%d Mbps;\n", Mbps); 
      response = buf;
      Logsystem::info_logger()->Tstamp() 
	<< "Setting Tx rate= "  << out_protocol[0]->Speed()*32 
	<< " Mbps.\n";
      //<< "Setting Tx rate= "  << out_protocol->Speed()*32 << " Mbps.\n";
      ret =0;
    }else{
      Logsystem::debug_logger()->Tstamp() << "Incorrecto cmd:" 
					  << str  << "\n"; 
      response = "!play_rate=0;\n";
      ret =-1;
    }

  }else if((pos = str.find("send_rate=")) != str.npos){
    string work = str.substr(pos+10);
    string::size_type tpos =  work.find_first_of(";:\n");
    int32_t Mbps = atoi(work.substr(0,tpos).c_str());
    //    int Mbps = rate/1000000;
    if( Mbps >0){   
      for(int i=0; i<ngrp; i++)
	out_protocol[i]->Set_Speed(1000000/32*Mbps, true);
      //	out_protocol[i].Set_Speed(1000000/32*Mbps, true);
      char buf[256]; sprintf(buf,"!send_rate=%d Mbps;\n", Mbps); 
      response = buf;
      ret =0;
    }else{
      response= "!send_rate=0;\n"; 
      ret =-1;
    }
    //    fprintf(stderr,"response=%s", response.c_str());
    sock->send((uint8_t*)response.c_str(),response.length());
  }else if((pos = str.find("net_protocol=")) != str.npos){
    string prto_str = str.substr(pos+13,prto_str.npos-13);
    pos = prto_str.find(":");
    string protocol = prto_str.substr(0,pos);
    prto_str = prto_str.substr(pos+1);
    pos = prto_str.find(":");
    int32_t bufsize  = strtol(prto_str.substr(0,pos).c_str(), 0, 10);
    fprintf(stderr,"protocol=%s, sockBuf=%d\n",protocol.c_str(), bufsize);
    if(bufsize <1){
      Logsystem::error_logger()->Tstamp()
	<<" E CmdParser(): Invalid bufsize by net_protocol command from Client\n";
      Logsystem::error_logger()->Tstamp() << "command=" << str << "\n";
      response= "!net_protocol = 0;\n";
      ret = -1;
    }

    if(protocol.find("UDP") != protocol.npos ||
       protocol.find("udp") != protocol.npos ){  
      //      k5vsi_sender->Sock_Create(PROTO_UDP);
      for(int i=0; i<ngrp; i++)
	out_protocol[i]->Set_Transprotocol(TransInfo::UDP_IO); 
      //	out_protocol[i].Set_Transprotocol(TransInfo::UDP_IO); 
    }
    if(protocol.find("TCP") != protocol.npos ||
       protocol.find("tcp") != protocol.npos ){ 
      //      k5vsi_sender->Sock_Create(PROTO_TCP); 
      for(int i=0; i<ngrp; i++)
	out_protocol[i]->Set_Transprotocol(TransInfo::TCP_IO); 
      //	out_protocol[i].Set_Transprotocol(TransInfo::TCP_IO); 
    }
        
    response= "!net_protocol=0;\n"; 
    ret = -1;
    //    fprintf(stderr,"response=%s", response.c_str());

  }else if((pos = str.find("in2net=connect:")) != str.npos){
    string work = str.substr(pos+15);
    string::size_type tpos =  work.find_first_of(":");
    if( tpos != work.npos ){ // host + port 
      string host_ip = work.substr(0,tpos);
      pos=  work.find_first_of(";\n");
      string sport = work.substr(tpos+1,pos-tpos-1);
      //      fprintf(stderr, "Host IP=%s port=%s\n", host_ip.c_str(),sport.c_str());
      Logsystem::info_logger()->Tstamp() << "Host IP=" << host_ip 
					 << " port= " << sport << "\n";
      int iport = strtol(sport.c_str(), 0, 10);
      for(int i=0; i<ngrp; i++) {// add 2014/12/9 miya
      out_protocol[i]->Set_Host(host_ip.c_str()); // modify 2014/12/9 miya 0->i
      out_protocol[i]->Set_DataPort(iport);} // modify 2014/12/9 miya 0->i
      //       out_protocol->Set_Host(host_ip.c_str());
      //       out_protocol->Set_DataPort(iport);
    }else{
      pos=  work.find_first_of(";\n");
      string host_ip = work.substr(0,pos);
      for(int i=0; i<ngrp; i++) {// add 2014/12/9 miya
      out_protocol[i]->Set_Host(host_ip.c_str());} // modify 2014/12/9 miya 0->i
      //      out_protocol->Set_Host(host_ip.c_str());
      Logsystem::info_logger()->Tstamp() << "Host IP=" << host_ip  << "\n";
      //      fprintf(stderr, "Host IP=%s\n", host_ip.c_str());
    }
    char buf[256];
    sprintf(buf,"!in2net=%s:%d\n", out_protocol[0]->Host().c_str(),
	    out_protocol[0]->DataPort());
    response= buf;
    ret =0;

  }else if((pos = str.find("in2net=on")) != str.npos){
    //    bool ok_flag=true;
    string stname;
    try{
      if(nout < 1){
	cerr << "E TransCtl::CmdParser(): nout <1." << endl;
	throw "E TransCtl::CmdParser(): nout <1. ";
      }
      
      
      for(int ith=0; ith < nout; ith++){
	if( out_protocol[ith]->TransProtocol() == TransInfo::FILE_IO) {
	  ret = dumper[0]->Set_Nout(nout);
	  for(int i=0; i < nout; i++){ 
	    ret = dumper[0]->Set_IO(out_protocol[i], indev, i);
	  } // add 2014/12/9 miya
	}else{
	  ret = dumper[ith]->Set_IO(out_protocol[ith], indev);	  
	}
	if( ret <0){
	  char buf[256];
	  sprintf(buf,"E Trans_ctl::CmdParser(): failed Dumper[%d]::Set_IO()",ith);
	  throw buf;
	  return -1;
	}
	  
//if( indev->DevType() == InDev_Model::K5VSI) comment 2014/12/9 miya
//ret = dumper[ith]->Set_IO(out_protocol[ith], (K5_VSI*)indev +ith); comment 2014/12/9 miya
// 	if((out_protocol[ith]->TransProtocol()== TransInfo::FILE_IO && ith==0) 
//    || (out_protocol[ith]->TransProtocol() != TransInfo::FILE_IO)){ // add 2014/12/9 miya
	//Logsystem::error_logger()->Tstamp() <<  "E Dumper::Set_IO()\n";

	sprintf(msg,"D TrancCtl::CmdParser(): dumper[%d]->Ready..\n",ith);
	Logsystem::debug_logger()->Tstamp()  << msg;
	Logsystem::debug_logger()->Flush();
	cerr << msg << endl;

	if(! dumper[ith]->Ready()){
	  throw "E Trans_ctl::CmdParser(): Dumper.Ready() failed."; 
	}
	sprintf(msg,"D TrancCtl::CmdParser(): dumper[%d]->Ready: Fnish\n",ith);
	cerr << msg << endl;
	Logsystem::debug_logger()->Tstamp() << msg;
	Logsystem::debug_logger()->Flush();
	
	dumper[ith]->Set_Span(-1);
	
	out_protocol[ith]->pVLBI_Frame()->Get_Stname(stname);
	cerr << "in2net=on,["<< ith+1 << "/" << nout << "] Stname = " 
	     << stname << endl;
	
 //     if(out_protocol[ith]->TransProtocol() != TransInfo::FILE_IO){ // add 2014/12/9 miya
	  Logsystem::debug_logger()->Tstamp() << "D TrancCtl::CmdParser(): dumper["
					      << ith << "]->start()..\n";
	  dumper[ith]->start();  // Starting sending/writing data 
	  sprintf(msg,"D TrancCtl::CmdParser(): dumper[%d]->start()ed.\n",ith);
	  Logsystem::debug_logger()->Tstamp() << msg;
	  Logsystem::debug_logger()->Flush();

	  if(out_protocol[0]->TransProtocol() == TransInfo::FILE_IO) ith=nout;
      } // for nout
      
//     // add 2014/12/9 miya start--------------------------------------------
//     if(out_protocol[0]->TransProtocol() == TransInfo::FILE_IO){
//   Logsystem::debug_logger()->Tstamp() << "D TrancCtl::CmdParser(): dumper["
// 					  << 0 << "]->start()..\n";
//       dumper[0]->start();  // Starting sending/writing data 
//   Logsystem::debug_logger()->Tstamp() << "D TrancCtl::CmdParser(): dumper["
// 					  << 0 << "]->start()ed. \n";
//       Logsystem::debug_logger()->Flush();		
//     }
//     // add 2014/12/9 miya end-------------------------------------------
  	
      string response= "!in2net=1;\n"; 
      sock->send((uint8_t *)response.c_str(),response.length());
    } // end try
    
    catch(const char* msg){
      Logsystem::error_logger()->Tstamp() 
	<< msg << "\n\t Stopping InptuDevice\n";
      string response= "!in2net=0;\n"; 
      cerr << msg << "\n\t Stop InputDevice \n";
      if(indev->DevType() == InDev_Model::K5VSI)
	for(int i=0; i<nout; i++) ((K5_VSI*)indev+i)->Stop();
      else     indev->Stop();
      
      sock->send((uint8_t *)response.c_str(),response.length());
    }
    
  }else if( str.find("in2net=disconnect") != str.npos ||
	    str.find("in2net=off") != str.npos  ||
	    str.find("play=off") != str.npos ){
	    for(int i=0; i < (int)nout; i++) {
	    	if((out_protocol[i]->TransProtocol() == TransInfo::FILE_IO && i == 0) || (out_protocol[i]->TransProtocol() != TransInfo::FILE_IO)){ // add 2014/12/9 miya
	    		dumper[i]->Stop();
	    	}
	    }
//    indev->Stop(); 
    string response= "!in2net=0;\n"; 
    //    fprintf(stderr,"response=%s", response.c_str());
    sock->send((uint8_t*)response.c_str(),response.length());
    Logsystem::info_logger()->Tstamp()
      << "I CmdParser(): in2net=off received. finishing.\n" ;
  }else  if(str.find("play=quit") != str.npos ){ // 2014/8/5 command add miya
  	if( indev->Status() == Thread::RUNNING ) { 
  		for(int i=0; i < (int)nout; i++) {
  			if((out_protocol[i]->TransProtocol() == TransInfo::FILE_IO && i == 0) || (out_protocol[i]->TransProtocol() != TransInfo::FILE_IO)){ // add 2014/12/9 miya
  			dumper[i]->Stop();}
  		}
  	}
	ctrl_state = TERM;
	//indev->Stop();
    string response= "!play=quit;\n"; 
    fprintf(stderr,"response=%s", response.c_str());
    sock->send((uint8_t*)response.c_str(),response.length());
    Logsystem::info_logger()->Tstamp()
      << "I CmdParser(): play=quit received. finishing.\n" ;
  }else  if(str.find("in2net?") != str.npos ){
    string response= "!in2net:connect=";
    for(int i=0; i<nout; i++){
      char buf[256];
      sprintf(buf,"%s:%d,", 
	      out_protocol[i]->Host().c_str(), out_protocol[i]->DataPort());
      response += buf;
    }
    response +=  ";\n";
    //    fprintf(stderr,"response=%s", response.c_str());
    ret =0;
    
  }else  if((pos = str.find("mode=")) != str.npos){ // single output
    vector <string> v;
    boost::algorithm::split(v, str, boost::is_any_of(":\n=;"));
    if( v.size() >= 4 ){
      int32_t srate = strtol( v[1].c_str(),NULL,10 )*1000000;
      int bit = strtol( v[2].c_str(),NULL,10 );
      int nch = strtol( v[3].c_str(),NULL,10 );
      const DAS_Info*  dasinf = indev->Das_inf();
      if(dasinf->Srate() < srate || dasinf->Nch() < nch
	 || dasinf->Qbit() < bit ){ // Wrong Cases 
	if( (pos = str.find("\n")) != str.npos){
	  str.replace(pos,1,"\0");
	}
	Logsystem::error_logger()->Tstamp()
	  << "E CmdParser():Requested mode:[" << str
	  << "] exceed inputMode:"  
	  << dasinf->Mode() << "\n";
	cerr << "E CmdParser():Requested mode:[" << str 
	     << "] exceed inputMode:"  
	     << dasinf->Mode() << "\n";
	response= "!mode=0;\n"; 
	ret = -1;
      }else{ // Mode input data is correct case 
	bool use_txspeed = false;
	DAS_Info das(srate, nch, bit);
	for(int igrp =0; igrp<ngrp; igrp++){
	  out_protocol[igrp]->Set_DasInf(das);
	  //	  out_protocol[igrp].pVLBI_Frame()->Set_Sparam(nch,bit,srate);
	  //	  const DAS_Info &das = out_protocol[igrp].pVLBI_Frame()->DAS_Inf();
	  cerr <<" Mode = "  <<  das.Mode() << endl;
	  if((out_protocol[igrp]->TransProtocol() == TransInfo::FILE_IO && igrp == 0) || (out_protocol[igrp]->TransProtocol() != TransInfo::FILE_IO)){ // add 2014/12/9 miya
	  	dumper[igrp]->SetDasMode(VirDumper::USR_OUT_SET); // external das setting  // modify 2015/3/9 miya Dumper->VirDumper
	  }
	  out_protocol[igrp]->Set_Speed(das.Speed(), use_txspeed); // use DASinf for txspeed
	  // 	  out_subgrp[igrp] = das;
	  // 	  out_subgrp[igrp].SetFirstCh( das.FirstCh() + nch*igrp );
	}
	char buf[256];
	sprintf(buf,"!mode=%dMsps:%dbit:%dch;\n", srate,bit,nch);
	nout=1;
	response= buf;
	ret = 0;
      }
    }else{
      fprintf(stderr,"E Parser(): incorrect command.\n");
      response= "!mode=0;\n"; 
      ret = -1;
    }
   
  }else  if((pos = str.find("modegrp=")) != str.npos){// Multiple out
    vector <string> v;
    boost::algorithm::split(v, str, boost::is_any_of(":\n=;"));
    if( v.size() >= 6 ){
      int igrp = strtol( v[1].c_str(),NULL,10 );
      int32_t srate = strtol( v[2].c_str(),NULL,10 )*1000000;
      int bit = strtol( v[3].c_str(),NULL,10 );
      int nch = strtol( v[4].c_str(),NULL,10 );
      int firstch = strtol( v[5].c_str(),NULL,10 );
      const DAS_Info*  dasinf = indev->Das_inf();
      if(dasinf->Srate() < srate || dasinf->Nch() < nch 
	 || dasinf->Qbit() < bit || firstch + nch > dasinf->Nch()
	 || igrp >= ngrp ){
	Logsystem::error_logger()->Tstamp()
	  << "E TransCtrl::CmdParser():Requested mode:" << str 
	  << " exceed inputMode:"   << dasinf->Mode() << "\n";
	cerr << "E Requested mode:" << str << "igrp=" << igrp 
	     << " exceed " << "ngrp="<< ngrp << "inputMode:" 
	     << dasinf->Mode() << "\n";
	response= "!mode=0;\n"; ret = -1;
      }else{
	DAS_Info das(srate, nch, bit, firstch);
	out_protocol[igrp]->Set_DasInf(das);
	//	out_protocol[igrp].pVLBI_Frame()->Set_DasInfo(das);
	fprintf(stderr," ModeGroup[%d] = %s\n ", igrp,das.Mode(3).c_str());
    if((out_protocol[igrp]->TransProtocol() == TransInfo::FILE_IO && igrp == 0) || (out_protocol[igrp]->TransProtocol() != TransInfo::FILE_IO)){ // add 2014/12/9 miya
		dumper[igrp]->SetDasMode(VirDumper::USR_OUT_SET); // external das setting  // modify 2015/3/9 miya Dumper->VirDumper
    }
	bool use_txspeed = false;
	out_protocol[igrp]->Set_Speed(das.Speed(), use_txspeed); 
	out_protocol[igrp]->ReAdjst_Nframe_sec(); // Added by M.S. 2014/11/12
	// 	  out_subgrp[igrp] = das;
	// 	  out_subgrp[igrp].SetFirstCh( das.FirstCh() + nch*igrp );
	char buf[256];
	sprintf(buf,"!modegrp=%d:%dMsps:%dbit:%dch:%dch;\n",
		igrp,srate/1000000,bit,nch,firstch); 
	response= buf;
	ret = 0;
      }
    }else{
      fprintf(stderr,"E Parser(): incorrect command.\n");
      response= "!modegrp=0;\n"; 
      ret = 0;
    }

  }else if((pos = str.find("mode?")) != str.npos){
    DAS_Info indas = *indev->Das_inf();
    response= "!mode= IN:" + indas.Mode() + "-OUT:";
    for( int igrp =0; igrp < nout; igrp++){
      const DAS_Info &outdas = out_protocol[igrp]->pVLBI_Frame()->DAS_Inf();
      response += outdas.Mode() + " ";     
    }
    response += ";\n";
    ret = 0;

  }else if((pos = str.find("DTS_id?")) != str.npos){
    response= "!DTS_id? 0:mark5b:-:-:-:-;\n"; 
    ret = 0;
  }else if((pos = str.find("host?")) != str.npos ||
	   (pos = str.find("destination?")) != str.npos ){
    response= "!destination=";
    for(int igrp=0; igrp< nout; igrp++){
      response += out_protocol[igrp]->Host() + " "; 
    }
    response += ";\n";
    ret = 0;

  }else if((pos = str.find("status?")) != str.npos){
    response= "!status?:";
    for(int igrp=0; igrp< nout; igrp++){
      response += out_protocol[igrp]->Status() + " ";
    }
    response += ";\n"; 
    ret = 0;

  }else if((pos = str.find("mtu=")) != str.npos){
    int tpos = str.find_last_of("1234567890",pos+4);
    string mtu_str = str.substr(pos+4, tpos-pos-3);
    int mtu= strtol(mtu_str.c_str(),NULL,10);
    if( mtu > 0 ){
      for(int igrp=0; igrp<ngrp; igrp++){
	if(mtu != out_protocol[igrp]->Mtu() ){
	  out_protocol[igrp]->Set_mtu(mtu);
	  out_protocol[igrp]->Create_sock(); // Re-createing socket 
	}
      }
      char buf[256];	
      sprintf(buf,"!mtu=%d;\n",mtu);
      response = buf;
      ret= 0;
    }else{
      Logsystem::error_logger()->Tstamp()
	<<  "E Trans_ctl::CmdParser(): negative MTU(" 
	<< mtu << ") \n";
      response = "!mtu=0;\n";
      ret = -1;
    }

  }else if((pos = str.find("mtu?")) != str.npos){
    string response = "!mtu=";
    for(int igrp=0; igrp < nout; igrp++){
      char tmp_buf[256];
      sprintf(tmp_buf,"%d ",out_protocol[igrp]->Mtu());
      response += tmp_buf;
    }
    response += ";\n";
    ret = 0;
    
  }else  if((pos = str.find("data_source?")) != str.npos){
    string response = "!data_source=";
    response += indev->DevType();  response += ";\n";
    ret = 0;

  }else  if((pos = str.find("data_source=")) != str.npos){
    string tmpsrc = str.substr(pos+16);
    vector <string> v;
    boost::algorithm::split(v, tmpsrc, boost::is_any_of(":\n=;"));
    if( ( pos = v[0].find("FILE")) != v[0].npos ){ // FILE
      int tpos(0);
      if( (tpos = v[0].find_last_of("0123456789")) == (int)v[0].npos ){ // Single
	//	in_subgrp[0].FileName() = v[1];
	response = "!data_source=FILE:" + v[1] + ";\n";
	ret = 0;
      }else{  // Multiple data source 
	tmpsrc = v[0].substr(pos+4,tpos);
	int igrp = atoi(tmpsrc.c_str());
	if(igrp < (int)nin){
	  //	  in_subgrp[igrp].FileName() = v[1];
	  char buf[256];
	  sprintf(buf,"!data_source=FILE(%d):%s;\n",igrp,v[1].c_str());
	  response = buf;
	  ret = 0;
	}else{
	  fprintf(stderr, 
		  "E data_source=FILE(%d)>Nin(%d), ", igrp, nin);
	  cerr << "Nin must be set by data_source=multi: command.\n"; 
	  response = "!data_source=FILE:0;\n";
	  ret =-1;
	}
      } // end of single/multi
    }else if( (pos = v[0].find("NET")) != v[0].npos ){ // NET
      int tpos(0);
      if( (tpos = v[0].find_last_of("0123456789")) == (int)v[0].npos ){ // Single
	//	in_subgrp[0].HostName() = v[1];
	//	in_subgrp[0].Port() = atoi(v[2].c_str());
	response = "!data_source=NET:" + v[1] + ";\n";
	ret = 0;
      }else{  // Multiple data source 
	tmpsrc = v[0].substr(pos+4,tpos);
	int igrp = atoi(tmpsrc.c_str());
	if(igrp < (int)nin){
	  //	  in_subgrp[igrp].HostName() = v[1];
	  //	  in_subgrp[igrp].Port() = atoi(v[2].c_str());
	  char buf[256];
	  sprintf(buf,"!data_source=NET(%d):%s;\n",igrp,v[1].c_str());
	  response = buf;
	  ret = 0;
	}else{
	  fprintf(stderr, 
		  "E data_source=NET(%d)>Nin(%d), ", igrp, nin);
	  cerr << "Nin must be set by data_source=multi: command.\n";
	  response = "!data_source=NET:0;\n";
	  ret = -1;
	}  // end of single/multi
      }
    }else if( ( pos = v[0].find("multi")) != v[0].npos ){ // multiple data source
      int ntmp = atoi(v[1].c_str());
      char buf[80];
      if(ntmp <= ngrp){
	nin = ntmp;
	sprintf(buf, "!data_source=multi:%d;\n",nin);
      }else{	
	strcpy(buf, "!data_source=multi:-1;\n");
      }
      response= buf;
      ret = 0;
    }else{
      cerr << "E Unexpected cmd: " << str << endl;
      response = "!data_source=-1;";
      ret = -1;
    }  // end of FILE/NET/mult for data_source
  }else  if((pos = str.find("data_out=FILE")) != str.npos){
    
    if(out_protocol[0]->TransProtocol() != TransInfo::FILE_IO){
      string msg="E Current VLBI-Protocol is not for FILE. Restart by '-O FILE' option\n";
      Logsystem::error_logger()->Tstamp() << msg << "\n";
      cerr << msg << endl;
      msg = "!data_out=-1";
      sock->send((uint8_t*)msg.c_str(),msg.length());  
      return -1;
    }
    
    string tmpsrc = str.substr( pos+9 );
    vector <string> v;
    boost::algorithm::split(v, tmpsrc, boost::is_any_of(":\n=;"));
    int tpos(0);
    if( (tpos = v[0].find_last_of("0123456789")) == (int)v[0].npos ){ // Single
      //	out_subgrp[0].FileName() = v[1];
      out_protocol[0]->Set_Host(v[1].c_str());
      response = "!data_out=FILE:" + v[1] + ";\n";
      ret = 0;
    }else{  // Multiple data output
      tmpsrc = v[0].substr(pos+4,tpos);
      int igrp = atoi(tmpsrc.c_str());
      if( igrp > nout ){ // Error out of range 
	char buf[256];
	sprintf(buf,"E data_out=FILE(%d)> Nout(%d)", igrp, nout);
	response = buf;
	Logsystem::error_logger()->Tstamp() 
	  << buf  << ", Nout must be set by data_out=multi: command.\n";
	cerr << buf << ", Nout must be set by data_out=multi: command."
	     << endl;
	Logsystem::info_logger()->Tstamp() << "response=" << response ;
	ret = -1;
      }
      bool unique_fname(true);
      for(int i=0; i<nout; i++) {
	if(i == igrp) continue;
	if(out_protocol[i]->Host() == v[1] ){ unique_fname = false; break;}
      }
      if( ! unique_fname){
	cerr << "data_out name("<< v[1] << ")is not unique\n"
	     << " Curent nemes are:";  
	for(int i=0; i<nout; i++) cerr << out_protocol[i]->Host() << " ";
	cerr << endl;
	Logsystem::error_logger()->Tstamp()
	  << "data_out name("<< v[1] << ")is not unique\n"
	  << " Curent nemes are:";  
	for(int i=0; i<nout; i++) 
	  *Logsystem::error_logger() << out_protocol[i]->Host() << " ";
	*Logsystem::error_logger() << "\n";
	response = "E Invalid Group Number";
	ret = -1;
      }
      
      out_protocol[igrp]->Set_Host(v[1].c_str());
      char buf[256];
      sprintf(buf,"!data_out=FILE(%d):%s;\n",igrp, v[1].c_str());
      response = buf;
      ret = 0;
    } // end of single/multi
  }else if( (pos = str.find("data_out=NET")) != str.npos ){ // Output to NET
    if(out_protocol[0]->TransProtocol() == TransInfo::FILE_IO){
      string msg="E Current VLBI-Protocol is not for NET.\n";
      Logsystem::error_logger()->Tstamp() << msg << "\n";
      cerr << msg << endl;
      msg = "!data_out=-1";
      sock->send((uint8_t*)msg.c_str(),msg.length());  
      return -1;
    }
    string tmpsrc = str.substr( pos+9 );
    vector <string> v;
    boost::algorithm::split(v, tmpsrc, boost::is_any_of(":\n=;"));
    int tpos(0);
    if( (tpos = v[0].find_last_of("0123456789")) == (int)v[0].npos ){ // Single
      int port = atoi(v[2].c_str());
      // 	out_subgrp[0].HostName() = v[1];
      // 	out_subgrp[0].Port() = port;
      out_protocol[0]->Set_Host(v[1].c_str());
      out_protocol[0]->Set_DataPort(port);
      response = "!data_out=NET:" + v[1] + ";\n";
      ret = 0;
    }else{  // Multiple data output
      tmpsrc = v[0].substr(pos+3,tpos);
      int igrp = atoi(tmpsrc.c_str());
      if(igrp < (int)nout){ // Normal case 
	int port = atoi(v[2].c_str());
	// 	  out_subgrp[igrp].HostName() = v[1];
	// 	  out_subgrp[igrp].Port() = port;
	out_protocol[igrp]->Set_Host(v[1].c_str());
	out_protocol[igrp]->Set_DataPort(port);
	char buf[256];
	sprintf(buf,"!data_out=NET(%d):%s;\n", igrp, v[1].c_str());
	response = buf;
	ret = 0;
      }else{
	fprintf(stderr, 
		"E data_out=NET(%d)>Nout(%d), ", igrp, nout);
	cerr << "Nout must be set by data_out=multi: command.\n";
	response = "!data_out=NET:-1;\n";
	ret = -1;
      }
    } // end of single/multi
  }else if( (pos = str.find("data_out=multi")) != str.npos ){ // NET
    string tmpsrc = str.substr( pos+9 );
    vector <string> v;
    boost::algorithm::split(v, tmpsrc, boost::is_any_of(":\n=;"));
    int ntmp = atoi(v[1].c_str());
    char buf[80];
    if(ntmp <= ngrp && ntmp > 0){
      nout = ntmp;
      sprintf(buf, "!data_out=multi:%d;\n",nout);
      response= buf;
      ret = 0;
    }else{
      strcpy(buf,"!data_out=multi:-1;\n");
      response= buf;
      ret = -1;
    }

  }else{
    response= "!Unexpected message;\n";
    Logsystem::info_logger()->Tstamp() << "response=" << response ;
    ret = -1;
  }
  sock->send((uint8_t*)response.c_str(),response.length());    
  return ret;  
  
}


bool TransCtl::SetupCheck(){
  bool ret=true;
  if(sock_ ==0){ 
    cerr <<"TransCtl::Check() error: socket is not ready.\n"; ret = false; 
  }
  //   if(vlbi_frame ==0){ 
  //       cerr <<"TransCtl::Check() error: vlbi_frame is not ready.\n";ret = false;  
  //   }
  //   if(das_info ==0){ 
  //     cerr <<"TransCtl::Check() error: das_info is not ready.\n";ret = false;}
  if(indev ==0){ 
    cerr <<"TransCtl::Check() error: input device is not ready.\n";ret = false;
  }
  if(out_protocol.size() ==0){ 
    cerr <<"TransCtl::Check() error: output protocol is not ready.\n";ret = false;
  }

  return ret;
}

void TransCtl::Set_InputDev(InDev_Model* inv){
  indev = inv;

  // Nch,Srate,.. User specified rather than raw input
  if(inv->DevType() == InDev_Model::K5VSI){
    for(int i=0; i < (int)dumper.size(); i++)
      dumper[i]->SetDasMode(VirDumper::USR_OUT_SET); // modify 2015/3/9 miya Dumper->VirDumper
  }
}
