/** 
 * @Version $Rev: 202 $
 * @Author M.Sekido
 * @Date  $Date: 2016-09-29 22:22:30 +0900 (Thu, 29 Sep 2016) $
 * $Id: VDIF_Header.cpp 202 2016-09-29 13:22:30Z vlbi $ 
 */

#include "VDIF_Header.hpp"

//======= VDIF libraries ==== from here ======


//time_t VDIF_Hdr::Ref_Epoch()const {
Time_dat VDIF_Hdr::Ref_Epoch()const {
  int iy = Ref_Epoch_hy_counter/2;
  int im = Ref_Epoch_hy_counter % 2;
  int mon = im ? 7 : 1;
  int year = iy +2000;

  Time_dat t1(year,mon,1,0,0,0);
  return t1;
}


/// Since Frame_Length have to be selected under the conditions:
///   -- Total data amount in a second have to be transfered by exact Integer number of packets.
///   -- Frame_Length must be expressed by 8Bytes units, 
int VDIF_Hdr::Set_Sparam(const int nch, const int qbit, const uint32_t Hz){
  Set_Qbit(qbit);
  Set_Nch(nch);
  Set_Srate(Hz/1000);
//   int32_t tsize = srate/8*nch*qbit; // Total Bytes/sec
//   int ps = 1500 - 32;
//   for(; ps>0; ps--){ if(tsize % ps == 0 && ps % 8 ==0) break; }
//   if(ps==0){
//     cerr << "Error in VDIF_Hdr::Set_Sparam(): Could not find appropriate packet size\n";
//     return -1;
//   }
  //  Frame_Length = (ps+32)/8 -1 ;
  return 0;
}

Time_dat VDIF_Hdr::Set_Epoch0(const time_t t1){

  time_t t2 = t1;
  struct tm tm1;
  //  , tm2;
  gmtime_r(&t2, &tm1);
//   memset(&tm2,0,sizeof(tm2));
//   tm2.tm_year = tm1.tm_year;
//   tm2.tm_mon = (tm1.tm_mon >= 6 ) ? 6 : 0;
//   tm2.tm_mday = 1;
  int year = tm1.tm_year + 1900;
  int mon = tm1.tm_mon + 1;
   
  //  time_t t0 = mktime(&tm2);
  Time_dat t0(year,mon,1,0,0,0);

  int dy  =tm1.tm_year - 100;
  int cnt = dy * 2;
  cnt += (tm1.tm_mon >= 6 ) ? 1 : 0;
  Ref_Epoch_hy_counter = cnt;
//   Time_dat t2(t1);
//   HY_Time h1 = t2.HY_Time(); 
//   Ref_Epoch_hy_counter = h1.Hyear_cnt();
  return t0;
}
// void VDIF_Hdr::Set_Epoch(const time_t t1, const time_t t0_ref){
//   Sec_from_Ref = t1 - t0_ref;
// }


void VDIF_Hdr::Set_Epoch(const Time_dat &t1){
  HY_Time h1 = t1.Hy_time();
  //  time_t t_work = t1.Time_t();

  //  double soy0 = Set_Epoch0(t_work).Soy();;
  //  double soy = t1.Soy();
  //  Sec_from_Ref =  soy - soy0;
  if( Ref_Epoch_hy_counter==0)
    Ref_Epoch_hy_counter = h1.Hyear_cnt();
  
  Sec_from_Ref = (int)(t1 - Ref_Epoch());
}

string VDIF_Hdr::Str_hd(const Prnt_t fmt){
  //  int v1 =Version_No, v2=exud_nict.ext_dat_ver;
  char strbuf[256]; 
  string retstr;

  switch(fmt){
  case CONST:
    // Print only constant
    sprintf(strbuf, "frm:%5d,Stn:%s:%s,Epoch:%s,ThredID=%d,Type:%s,Ref:%s,FrmLen=%d ",
	    iFrame(), Stname().c_str(), Stnid().c_str(),
	    Epoch().Print_time(21).c_str(),
	    Thread_Id, Dtype().c_str(), Ref_Epoch().Print_time(21).c_str(),
	    Framesize()); 
    retstr = strbuf;
    if(EDV() == EDV_NICT){// NICT's extended user data
      if(Srate() < 1000000)
	sprintf(strbuf,"Mode:%dksps/%dbit/%dch ",  Srate()/1000,Qbit(), Nch());
      else
	sprintf(strbuf,"Mode:%dMsps/%dbit/%dch ",  Srate()/1000000,Qbit(), Nch());
    }else{
      sprintf(strbuf,"Mode:%dbit/%dch ",  Qbit(), Nch());
    }
    retstr += strbuf;
    break;
  case VAR:
    { // Print variable part by frames
      if( EDV() == EDV_NICT){ // NICT's extended user data
	double dt = (double)(Framesize() - Headsize())*8./Qbit()/Nch()/Srate();
	Time_dat t1 = Epoch() + dt * iFrame();
	sprintf(strbuf,"ifrm=%2d Epoch=%s", 
		iFrame(),t1.Print_time(20).c_str());
	retstr = strbuf;
      }else{
	sprintf(strbuf,"ifrm=%2d Epoch=%s", 
		iFrame(),Epoch().Print_time(21).c_str());
	retstr = strbuf;
      }
      if(Invalid_flag){  sprintf(strbuf," Invalid:");}
      else{              sprintf(strbuf," Valid  :");  }
      retstr += strbuf;
//       sprintf(strbuf," Ref=%s", Ref_Epoch().Print_time(21).c_str());      
//       retstr += strbuf;
      
    }
    break;
  case ONELINE:
    { // Print variable part by frames without CR
      string ep = Epoch().Print_time(15);
      sprintf(strbuf,"ifrm=%5d Epoch=%s FrmLen=%6d", 
	      iFrame(), ep.c_str(), Framesize());
      retstr = strbuf;
      if(Invalid_flag){  sprintf(strbuf," Invalid:");}
      else{              sprintf(strbuf," Valid  :");  }
      retstr += strbuf;

//       string ep0 = Ref_Epoch().Print_time(21);
//       fprintf(fp," Ref=%s", ep0.c_str());      
      if(EDV() == EDV_NICT){ // NICT's extended user data
	if(Srate() < 1000000)
	  sprintf(strbuf," %3dksps/%dbit/%dch ",Srate()/1000,Qbit(), Nch());
	else
	  sprintf(strbuf,"Mode: %3dMsps/%dbit/%dch ",Srate()/1000000,Qbit(), Nch());
      }else{
	sprintf(strbuf," %dbit/%dch ", Qbit(), Nch());
      }
      retstr += strbuf;
    }
    break;
  case ALL_ONE: // Print everything in one line
    sprintf(strbuf," %5ld %s %d %8s:%2s %3d %1d %2d %d %s %s %s",
	    (long)iFrame(), Epoch().Print_time(10).c_str(), 
	    ThreadID(), Stname().c_str(),Stnid().c_str(),
	    Srate()/1000000, Qbit(),Nch(), Framesize(),
	    Version().c_str(), Dtype().c_str(),
	    Ref_Epoch().Print_time(21).c_str());
    retstr = strbuf;
    break;
  case ALL:
  default:{
    //    sprintf(strbuf, "-VDIFHD (Ver. %d.%d):", v1,v2);
    sprintf(strbuf, " Frm:%5d ", iFrame());
    retstr = strbuf;
    if(Invalid_flag){  sprintf(strbuf," Data:Invalid ");}
    else{              sprintf(strbuf," Data:Valid ");  }
    retstr += strbuf;
   
    string ep0 = Ref_Epoch().Print_time(21);
    string ep = Epoch().Print_time(20);
    sprintf(strbuf,"Ref=%s Epoch=%s\n", ep0.c_str(),ep.c_str());
    retstr += strbuf;

    //    int fl = (Frame_Length)*8;
    sprintf(strbuf, "\t FrmSize=%5d Byte", Framesize());
    retstr += strbuf;

    string dat_type = Complex_data_flag ? "Cmplx" : "Real";
    sprintf(strbuf," StnID:%c%c Thread=%d Type:%s",
	    Station_ID[0], Station_ID[1], 
	    Thread_Id, dat_type.c_str());
    retstr += strbuf;
    
    uint8_t iedv = EDV();
    if(iedv == EDV_NICT || iedv == EDV_NPDF){ // NICT's extended user data
      if(Srate() < 1000000)
	sprintf(strbuf," Mode:%3dksps/%dbit/%dch ",  Srate()/1000,Qbit(), Nch());
      else
	sprintf(strbuf," Mode:%3dMsps/%dbit/%dch ",  Srate()/1000000,Qbit(), Nch());
      retstr += strbuf;

      switch(iedv){
      case  EDV_NICT:
	{
	  string stname;
	  exud_ver1.Get_Stname(stname);
	  sprintf(strbuf," Stn:%s ",stname.c_str());
	  retstr += strbuf;
	}
	break;
      case EDV_NPDF : 
	{
	  VDIF_EUD_npdf* pedv = static_cast <VDIF_EUD_npdf *> (&exud_ver1);
	  sprintf(strbuf," RF LowBnd:%f MHz ", pedv->RFlb()*1.e-3);
	  retstr += strbuf;
	}
      default: break;
      }

    }
  }
    break;
  }
  
  return retstr;
}


const bool VDIF_Hdr_find(const uint8_t*  buf, const uint32_t nbyte, 
			 long& offset, VDIF_Hdr *vdif_hdr){
  uint32_t remain = nbyte;
  uint8_t *pb = (uint8_t* )buf;

  VDIF_Hdr *pt_head(0); 
  uint32_t hdsize = pt_head->Headsize();
  int32_t count=0;

  while( remain >= hdsize ){
    pt_head = (VDIF_Hdr*)pb;
    switch( pt_head->EDV()){
    case EDV_NICT:
    case EDV_NPDF:
    case EDV_NRAO:
      if( pt_head->HD_Match() ){
	*vdif_hdr = *pt_head;
	offset =  count;
	return( true);
      }
      break;
    case EDV_STD:// Standard VDIF 
	*vdif_hdr = *pt_head;
	offset = count;
	return(true);
	break;
    default : // unlikely correct VDIF header.
	count++; remain--; pb++;

      //    }else{// Standard VDIF 
      //      if(pt_head->Framesize() < 1000 || 
      //	 pt_head->exud_nict.ext_dat_ver > NICT){
	//      }else{ // May be correct VDIF header 
	break;
    } // End of SW
    //  } // end for
  } // end while 
  
  return(false);
}

// Return offset and vdif_head
const bool VDIF_Hdr_find(FILE *fp, long &offset,  VDIF_Hdr *vdif_head){

  VDIF_Hdr*  pt_head(0);
  int32_t count=0;
  int32_t hdsize = pt_head->Headsize();
  int bufsize = hdsize * 10;
  uint8_t buf[bufsize];
  int nerr =0;
  try{
    while( feof(fp) == 0 && nerr < 3){
      int nread = fread(buf,bufsize,1,fp);
      if( nread <1 ){
	throw "VDIF_Hdr_find fread() returns 0";
      }
      pt_head  = (VDIF_Hdr*)buf;
    
      for(int i=0; i<bufsize; i++){
	//    if(dbg_disp_level>10) pt_head->Disp();
	switch(pt_head->EDV()){
	case EDV_NICT: // NICT Extension
	case EDV_NPDF: // NICT Extension
	case EDV_NRAO: // NRAO Extension
	  if(pt_head->HD_Match() ){
	    fseek(fp,count-bufsize,SEEK_CUR);
	    *vdif_head = *pt_head;
	    offset = count;
	    return(true);
	  }; break;
	case EDV_STD: // Standard VDIF 
	  //	case EDV_NON: // Standard VDIF 
	case EDV_HAYSTACK: // Standard VDIF 
	  if(pt_head->Framesize() < 1000 || 
	     (pt_head->Sec_Ref() ==0 && pt_head->Ref_Epoch_hy_counter ==0)
	     ){
	    // unlikely correct VDIF header.
	    count++; 
	    pt_head = (VDIF_Hdr*)(buf + i);
	  }else{ // May be correct VDIF header 
	    fseek(fp,count-bufsize,SEEK_CUR);
	    *vdif_head = *pt_head;
	    offset = count;
	    return(true);
	  }
	  break;
	} // End of Switch
      } // End for 
      nerr++;
    } // end while 
} // end try

  catch(const char *msg){
    cerr << msg << strerror(errno) << endl;
    if(feof(fp)) cerr << "EOF found.\n";
  }

  return (false);

}


/// ifmt=0: yyyy/mon/day-hh:mm:ss
// string VDIF_Hdr::Str_Ref_Epoch(const int ifmt){
  
//   string str;
//   time_t t1 = Ref_Epoch();
//   struct tm tm1;
//   gmtime_r(&t1, &tm1);
//   char buf[80];
//   switch(ifmt){
//   case 0:
//     strftime(buf,80,"%Y/%m/%d-%H:%M:%S", &tm1);
//     str = buf;
//     break;
//   case 1:
//   default:
//     fprintf(stderr,"VDIF_Hdr::Str_Ref_Epoch(): ifmt(%d) is not implemented\n",
// 	    ifmt);
//     break;
//   }

//   return str;
// }
//======= VDIF libraries ==== to   here ======
