/** 
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (水, 03  7月 2013) $
 * $Id: vdif.hpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>

extern "C"{
  #include <stdint.h>
}


/** VDIF header class */
class VDIF_header {
  /** Secondsn from reference Epoch **/
  unsigned Sec_from_Ref:30;
  /** Flag for legacy compatibility, 0:32 Bytes header, 1:16 Bytes header*/
  unsigned  Legacy_flag:1;
  /** Invalid flag 1: Invalid, 0: Valid */
  unsigned  Invalid_flag:1;
  /** Data Frame Number */
  unsigned  Frame_Number:24;
  /** Reference Epoch half year counter since 2000.0 */
  unsigned  Ref_Epoch_hy_counter:6;
  /** unused  */
  unsigned  unused:2;
  /** Data Frame length -1, unit: 8 bytes  */
  unsigned  Frame_Length:24;
  /** Log2(nch)*/
  unsigned  Lg_Nch:5;
  /** VDIF Version Number */
  unsigned  Version_No:3;
  /** Station ID*/
  unsigned char Station_ID[2];
  /** Thread ID*/
  unsigned  Thread_Id:10;
  /** Quantization Bit/sample -1 */
  unsigned  Bit:5;
  /** Comlex data flag. 1:Complex, 0: Real data */
  unsigned  Complex_data_flag:1;
  /** Extended user data 1 */
  uint8_t user_data1[3];
  /** Extended Data Version */
  uint8_t  Extended_Data_Version;
  /** Extended user data 2 */
  uint8_t user_data2[12];
public:
  VDIF_header():Sec_from_Ref(0),Legacy_flag(0),Invalid_flag(0),
  Frame_Number(0),Ref_Epoch_hy_counter(0),unused(0),
		Frame_Length(186),Lg_Nch(0),Version_No(1),Thread_Id(0),
		Bit(1),Complex_data_flag(0),Extended_Data_Version(0)
  {strncpy((char*)Station_ID,"Kb",2); memset(user_data1,0,3); memset(user_data2,0,12);}
  const uint32_t add(void){Sec_from_Ref++; return (uint32_t)Sec_from_Ref;}
  const uint32_t Frame_up(void){Frame_Number++; return (uint32_t)Frame_Number;}
  void Set_Invalid(){Invalid_flag=1;}
  void Set_SecRef(const uint32_t sec){Sec_from_Ref=sec;}
  void Print(FILE *fp=stdout);
};

inline void VDIF_header::Print(FILE *fp){
  char *pvalid(0), *plegacy;
  if(Invalid_flag) pvalid = "InValid"; else pvalid = "Valid";
  if(Legacy_flag) plegacy = "Legacy Mode"; else plegacy = "VDIF Normal";
  fprintf(fp,"**** VDIF header **** \n");
  fprintf(fp,"V:%s L:%s ", pvalid, plegacy);
  fprintf(fp,"Epoch: %ld sec from %.1lf\n", 
	  Sec_from_Ref, 2000.+Ref_Epoch_hy_counter*0.5);
  fprintf(fp,"Frame Num = %ld , size: %ld bytes, ", 
	  Frame_Number, (Frame_Length+1)*8);
  fprintf(fp,"Thread ID=%d, StationID: %c%c \n", 
	  Thread_Id, Station_ID[0], Station_ID[1]);
  fprintf(fp,"Nch=%d, Bit:%d\n", 1 << Lg_Nch, Bit);
  fprintf(fp,"************************************\n");
}
